import xml.etree.ElementTree as ET
from collections import OrderedDict
import numpy as np

try:
    import networkx as nx
    NX = True
except ImportError:
    prNX = False


class EDM:
    def __init__(self, filename):
        """
        Initiate an instance of an EDM object.

        Parameters
        ----------
            filename: str
                The path and filename of the EDM file to be imported.
        """
        self.tree = ET.parse(filename)
        self.root = self.tree.getroot()
        self._wellbore_id_to_name()
        self._wellbore_id_to_well_id()

    def get_attributes(self, tags=None, attributes={}, logic='AND'):
        """
        Get the attributes for the given tags in an EDM instance.

        Parameters
        ----------
            tags: str or list of str (default: None)
                The tag or list of tags you wish to return. The default will
                return all the tags that satisfy the given attributes.
            attributes: dict
                A dictionary of attribute keys and values to satisfy in the
                search of tags.
            logic: str (default: 'AND')
                Indicates whether the attributes should be all be satisfied
                ('AND') or if only one needs to be satisfied ('OR').

        Returns
        -------
            data: dict
                A dictionary of a list of dictionaries of tags and their
                attributes.
        """
        assert logic in ['AND', 'OR'], "logic must be 'AND' or 'OR'"
        if tags is None:
            tags = self.get_tags()
        data = {}
        if not isinstance(tags, list):
            tags = [tags]
        for child in self.root:
            if child.tag in tags:
                if bool(attributes):
                    booleans = []
                    for k, v in attributes.items():
                        if not isinstance(v, list):
                            v = [v]
                        if k in child.attrib and child.attrib[k] in v:
                            booleans.append(True)
                        else:
                            booleans.append(False)
                        if bool(booleans):
                            if (
                                logic == 'AND' and all(booleans)
                                or logic == 'OR' and any(booleans)
                            ):
                                try:
                                    data[child.tag].append(child.attrib)
                                except KeyError:
                                    data[child.tag] = [child.attrib]
                else:
                    try:
                        data[child.tag].append(child.attrib)
                    except KeyError:
                        data[child.tag] = [child.attrib]
        return data

    def _wellbore_id_to_name(self):
        """
        It's easier for the user to reference the name of the wellbore rather
        thatn the id, so let's make a couple of dictionaries to quickly
        perform lookups.
        """
        self.wellbore_id_to_name = {}
        self.wellbore_name_to_id = {}
        for _, v in self.get_wellbore_ids().items():
            self.wellbore_id_to_name[v['wellbore_id']] = v['wellbore_name']
            self.wellbore_name_to_id[v['wellbore_name']] = v['wellbore_id']

    def get_wells(self):
        return {
            f"{child.attrib['well_common_name']}": child.attrib
            for child in self.root
            if child.tag == 'CD_WELL_ALL'
        }

    def get_sites(self):
        return {
            f"{child.attrib['site_name']}": child.attrib
            for child in self.root
            if child.tag == 'CD_SITE'
        }

    def get_wellbore_ids(self):
        return {
            f"{child.attrib['wellbore_name']}": child.attrib
            for child in self.root
            if child.tag == 'CD_WELLBORE'
        }

    def _wellbore_id_to_well_id(self):
        self.well_id_to_wellbore_id = {}
        for child in self.root:
            if child.tag == 'CD_WELLBORE':
                self.well_id_to_wellbore_id[child.attrib['wellbore_id']] = (
                    child.attrib['well_id']
                )

    def get_tags(self, sort=True):
        tags = list(set([child.tag for child in self.root]))
        if sort:
            return sorted(tags)
        else:
            return tags

    def get_wellbore_graph(self):
        assert NX, "ImportError: try pip install welleng[all]"
        # get nodes
        G = nx.Graph()

        for _, v in self.get_wellbore_ids().items():
            if 'parent_wellbore_id' in v:
                G.add_edge(v['parent_wellbore_id'], v['wellbore_id'])
            else:
                G.add_node(v['wellbore_id'])

        return G

    def add_attributes(self, attributes, additional):
        if bool(additional):
            for k, v in additional.items():
                try:
                    attributes[k].extend(v)
                except KeyError:
                    attributes[k] = v
        return attributes

    def get_wellbore_data(self, wellbore_id):
        attributes = self._get_data(
            attributes={}, attr="wellbore_id", attr_id=wellbore_id
        )
        # some additional attributes are needed:
        well_id = attributes['CD_WELLBORE'][0]['well_id']
        tags = [
            'CD_WELL', 'CD_DATUM'
        ]
        additional = self.get_attributes(
            tags=tags, attributes={'well_id': well_id}
        )
        attributes = self.add_attributes(attributes, additional)

        # for the surveys, need to extract survey data from the parent wells
        predecessors = self.get_parents(wellbore_id)
        if bool(predecessors):
            tags = ['CD_SURVEY_HEADER', 'CD_SURVEY_STATION']
            additional = self.get_attributes(
                tags=tags, attributes={'wellbore_id': predecessors}
            )
        attributes = self.add_attributes(attributes, additional)

        return attributes

    def get_parents(self, wellbore_id, predecessors=[]):
        for child in self.root:
            if (
                child.tag == "CD_WELLBORE"
                and child.attrib['wellbore_id'] == wellbore_id
            ):
                if 'parent_wellbore_id' in child.attrib:
                    parent = child.attrib['parent_wellbore_id']
                    predecessors.append(parent)
                    predecessors = self.get_parents(parent, predecessors)
                else:
                    return predecessors
        return predecessors

    def _get_data(self, attributes, attr, attr_id):
        # get list of tags
        tags = self.get_tags()
        attributes = self.get_attributes(tags, attributes={attr: attr_id})

        return attributes

    def get_wellbore(self, wellbore, name=False):
        if name:
            wellbore_name = wellbore
            # look-up well_id in "CD_WELLS"
            wellbore_id = self.wellbore_name_to_id[wellbore_name]
        else:
            wellbore_id = wellbore
            wellbore_name = self.wellbore_id_to_name[wellbore_id]

        well = Well(
            wellbore_id,
            wellbore_name,
            self.get_wellbore_data(wellbore_id)
        )        

        return well

    def get_case_name_from_id(self, case_id):
        for child in self.root:
            if (
                child.tag == 'CD_CASE'
                and child.attrib['case_id'] == case_id
            ):
                return child.attrib['case_name']


class Case(object):
    def __init__(self, case):
        for a, b in case.items():
            if isinstance(b, (list, tuple)):
                setattr(
                    self, a, [Case(x) if isinstance(x, dict) else x for x in b]
                )
            else:
                setattr(self, a, Case(b) if isinstance(b, dict) else b)


class Well:
    def __init__(self, wellbore_id, wellbore_name, well_data):
        self.wellbore_id = wellbore_id
        self.wellbore_name = wellbore_name
        self.well_data = well_data

    # def get_scenarios(self):
    #     if 'CD_SCENARIO' in self.well_data:
    #         scenarios = [
    #             (s['scenario_id'], s['name'])
    #             for s in self.well_data['CD_SCENARIO']
    #         ]
    #     else:
    #         scenarios = None
    #     return scenarios

    def _get_cases(self):
        return {
            c['case_id']: c['case_name']
            for c in self.well_data['CD_CASE']
        }

    def _set_attribute(
        self, case, label, tag_label=None, prefix="CD", suffix="id",
        out_label=None, **kwargs
    ):
        if tag_label is None:
            tag_label = label
        if out_label is None:
            out_label = label
        if hasattr(case, f"{label}_{suffix}"):
            setattr(
                case,
                out_label,
                self.get_data(
                    getattr(
                        case, f"{label}_{suffix}"
                    ),
                    label=label,
                    tag_label=tag_label,
                    prefix=prefix,
                    suffix=suffix
                )
            )
        return case

    def make_case(self, case_id):
        case = None
        for row in self.well_data['CD_CASE']:
            if row['case_id'] == case_id:
                case = Case(row)
        if case is not None:
            # the order of the list is important
            attributes = [
                {'label': 'wellbore'},
                {'label': 'fluid'},
                {'label': 'scenario'},
                {'label': 'well'},
                {'label': 'datum'},
                {'label': 'hole_sect_group'},
                {'label': 'assembly'},
                {'label': 'hole_sect', 'suffix': "group_id"},
                {'label': 'assembly', 'tag_label': 'assembly_comp'},
                {
                    'label': 'wellbore', 'tag_label': 'survey_header',
                    'out_label': 'survey_header'
                }
            ]
            # TODO could probably do this is a single loop?
            for k in attributes:
                case = self._set_attribute(
                    case, **k
                )
        self._get_surveys(case)
        self._get_ppfpt(case)

        return case

    def _get_ppfpt(self, case):
        key_words = ['PORE', 'FRAC', 'CD_TEMP']
        anti_key_words = ['TEMPLATE']
        for t in self.well_data:
            for kw in key_words:
                if kw in t:
                    for akw in anti_key_words:
                        if akw not in t:
                            data = self._sort_ppfpt_data(self.well_data[t], kw)
                            setattr(
                                case,
                                t.lower()[3:],
                                data
                                # self.well_data[t]
                            )
    @staticmethod
    def _sort_ppfpt_data(data, kw):
        lookup = {
            'PORE': 'pore_pressure',
            'FRAC': 'frac_gradient_pressure',
            'CD_TEMP': 'temperature'
        }
        kw = lookup[kw]
        sorted_data = {}
        sorted_sorted_data = {}
        for d in data:
            key = [k for k in d if "_group_id" in k][0]
            try:
                sorted_data[d[key]].append(d)
            except KeyError:
                sorted_data[d[key]] = [d]
        for item in sorted_data.items():
            try:
                sorted_sorted_data[item[0]] = sorted(
                    item[1], key=lambda k: float(k['tvd'])
                )
            except KeyError:
                return sorted_data
        return sorted_sorted_data

    def _get_surveys(self, case):
        # get the surveys... turns out to be a bit of a PITA
        if hasattr(case, "survey_header"):
            survey_header_ids = [
                h['survey_header_id']
                for h in case.survey_header
            ]
            case.survey_station = {}
            for sh in survey_header_ids:
                case.survey_station[sh] = [
                    s for s in self.well_data['CD_SURVEY_STATION']
                    if s['survey_header_id'] == sh
                ]
            md = {
                sh['survey_header_id']: float(sh['md_max'])
                for sh in case.survey_header
            }
            survey_lists = {}
            for sh in survey_header_ids:
                survey_lists[sh] = self.get_parent_surveys(
                    sh,
                    data=OrderedDict({sh: md[sh]})
                )
            survey = {}
            for item in survey_lists.items():
                is_first = True
                for k, v in item[1].items():
                    if is_first:
                        survey[item[0]] = []
                        # current_survey_id = k
                        current_md = md[k]
                        is_first = False
                    current_survey_id = k
                    if float(v) < current_md:
                        for station in self.well_data['CD_SURVEY_STATION']:
                            if (
                                station['survey_header_id'] == current_survey_id
                                and float(station['md']) >= float(v)
                                and float(station['md']) <= current_md 
                            ):
                                try:
                                    survey[item[0]].append(station)
                                except KeyError:
                                    survey[item[0]] = [station]
                        current_survey_id = k
                        current_md = float(v)
                survey[item[0]] = sorted(survey[item[0]], key=lambda k: float(k['md']))
        case.survey = survey

    def get_parent_surveys(self, survey_header_id, data=OrderedDict()):
        for sh in self.well_data['CD_SURVEY_HEADER']:
            if (
                sh['survey_header_id'] == survey_header_id
                and 'tie_survey_header_id' in sh
            ):
                tsh_id = sh['tie_survey_header_id']
                data[survey_header_id] = sh['tie_on_depth']
                self.get_parent_surveys(tsh_id, data)
        return data

    def get_data(
        self, label_id, label, tag_label=None, prefix="CD", suffix="id"
    ):
        if tag_label is None:
            tag_label = label
        t = f"{prefix}_{tag_label}".upper()
        k = f"{label}_{suffix}"
        data = []
        for row in self.well_data[t]:
            if row[k] == label_id:
                data.append(row)
        if bool(data):
            return data
        else:
            return None

    def get_hole_sections(self):
        hole_sect_group_ids = []
        for k, v in self.well_data.items():
            hole_sect_group_ids.extend(
                [
                    p for row in v
                    for t, p in row.items()
                    if t == "hole_sect_group_id"
                ]
            )
        return list(set(hole_sect_group_ids))

    def _get_hole_section_data(self, group_id):
        hole_section_data = {}
        for k, v in self.well_data.items():
            for row in v:
                if (
                    "hole_sect_group_id" in row
                    and row["hole_sect_group_id"] == group_id
                ):
                    hole_section_data[k] = {}
                    for t, p in row.items():
                        hole_section_data[k][t] = p
        return hole_section_data

    def get_hole_section_data(self, hole_sections=None):
        """
        Parameters
        ----------
            hole_sections: list of str
        """
        hole_section_data = {}
        if hole_sections is None:
            hole_sections = self.get_hole_sections()
        for hs in hole_sections:
            hole_section_data[hs] = self._get_hole_section_data(hs)
        return hole_section_data

    def get_ppfpt_data(self):
        try:
            pp = self._get_ppfpt_data(
                self.well_data['CD_PORE_PRESSURE'],
                'pore_pressure_group_id',
                'pore_pressure'
            )
        except KeyError:
            pp = None
        try:
            fp = self._get_ppfpt_data(
                self.well_data['CD_FRAC_GRADIENT'],
                'frac_gradient_group_id',
                'frac_gradient_pressure'
            )
        except KeyError:
            fp = None
        try:
            tp = self._get_ppfpt_data(
                self.well_data['CD_TEMP_GRADIENT'],
                'temp_gradient_group_id',
                'temperature'
            )
        except KeyError:
            tp = None

        return {
            'pore_pressure': pp,
            'fracture_pressure': fp,
            'temperature': tp
        }

    @staticmethod
    def _get_ppfpt_data(data, *args):
        """
        Helper function for extracting tvd linked data.

        Parameters
        ----------
            data: extracted EDM arrtibute data
            *args: arg1, arg2
                arg1: string
                    key name for the data's group_id
                arg2: string
                    key_name for the required data

        Returns
        -------
            dict of (n, 2) arrays
        """
        arg1, arg2 = args
        temp = {}
        for p in data:
            if p[arg1] in temp:
                temp[p[arg1]].extend([
                    [
                        float(p['tvd']),
                        float(p[arg2])
                    ]
                ])
            else:
                temp[p[arg1]] = ([
                    [
                        float(p['tvd']),
                        float(p[arg2])
                    ]
                ])
        for group in temp:
            temp[group] = np.vstack(
                sorted(temp[group], key=lambda tup: tup[0])
            ).reshape(-1, 2)

        return temp


if __name__ == "__main__":
    # import os
    # os.environ['DISPLAY'] = ':1'

    # import EDM data
    FILENAME = 'data/Volve.xml'
    edm = EDM(FILENAME)

    well_ids = list(edm.get_wells())
    wb_ids = list(edm.get_wellbore_ids().keys())

    # name of the wellbore of interest
    well_name = well_ids[0]
    wellbore_name = wb_ids[0]

    # extract data for given wellbore name
    well = edm.get_wellbore(wellbore_name, name=True)

    case = well.make_case('3jcyt')

    hole_sections = well.get_hole_sections()
    hole_section_data = well.get_hole_section_data(hole_sections)

    G = edm.get_wellbore_graph()

    print("Done")



import numpy as np
from .utils import (
    get_unit_vec, get_vec, get_angles, get_nev, get_xyz,
)
# from .connector import Connector


class Node:
    def __init__(
        self,
        pos=None,
        vec=None,
        md=None,
        inc=None,
        azi=None,
        unit='meters',
        degrees=True,
        nev=True,
        cov_nev=None,
        **kwargs
    ):
        self.check_angle_inputs(inc, azi, vec, nev, degrees)
        self.get_pos(pos, nev)
        self.md = md
        self.unit = unit
        self.cov_nev = np.zeros(shape=(1, 3, 3)) if cov_nev is None else cov_nev

        for k, v in kwargs.items():
            setattr(self, k, v)

    def check_angle_inputs(self, inc, azi, vec, nev, degrees):
        if all(v is None for v in [inc, azi, vec]):
            self.vec_xyz = None
            self.vec_nev = None
            self.inc_rad = None
            self.inc_deg = None
            self.azi_rad = None
            self.azi_rad = None
            return
        elif vec is None:
            assert inc is not None and azi is not None, (
                "vec or (inc, azi) must not be None"
            )
            self.vec_xyz = get_vec(inc, azi, deg=degrees).reshape(3).tolist()
            self.vec_nev = get_nev(self.vec_xyz).reshape(3).tolist()
        else:
            if nev:
                self.vec_nev = get_unit_vec(vec).reshape(3).tolist()
                self.vec_xyz = get_xyz(self.vec_nev).reshape(3).tolist()
            else:
                self.vec_xyz = get_unit_vec(vec).reshape(3).tolist()
                self.vec_nev = get_nev(self.vec_xyz).reshape(3).tolist()
        self.inc_rad, self.azi_rad = (
            get_angles(self.vec_xyz).T
        ).reshape(2)
        self.inc_deg, self.azi_deg = (
            np.degrees(np.array([self.inc_rad, self.azi_rad]))
        ).reshape(2)

    def get_pos(self, pos, nev):
        if pos is None:
            self.pos_xyz = None
            self.pos_nev = None
            return
        if nev:
            self.pos_nev = np.array(pos).reshape(3).tolist()
            self.pos_xyz = get_xyz(pos).reshape(3).tolist()
        else:
            self.pos_xyz = np.array(pos).reshape(3).tolist()
            self.pos_nev = get_nev(pos).reshape(3).tolist()

    @staticmethod
    def get_unit_vec(vec):
        vec = vec / np.linalg.norm(vec)

        return vec

    def properties(self):
        return vars(self)


def get_node_params(node):
    pos = node.pos_nev
    vec = node.vec_nev
    md = node.md

    return pos, vec, md



import numpy as np
from numpy import sin, cos, tan, pi, sqrt
from numpy.core.defchararray import index
import yaml
import os
from collections import OrderedDict
# import imp

# import welleng.error
from ..utils import NEV_to_HLA

# since this is running on different OS flavors
PATH = os.path.dirname(__file__)
TOOL_INDEX = os.path.join(
    '', *[PATH, 'tool_index.yaml']
)

ACCURACY = 1e-6


class ToolError:
    def __init__(
        self,
        error,
        model
    ):
        """
        Class using the ISCWSA listed tool errors to determine well bore
        uncertainty.

        Parameters
        ----------
        error: an intitiated welleng.error.ErrorModel object
        model: string

        Returns
        -------
            errors: welleng.error.ErrorModel object
                A populated ErrorModel object for the selected error model.
        """
        error.__init__

        self.e = error
        self.errors = {}

        filename = os.path.join(
            '', *[PATH, 'tool_codes', f"{model}.yaml"]
        )

        with open(filename, 'r') as file:
            self.em = yaml.safe_load(file)

        # for gyro tools the continuous survey errors need to be done last
        self.em['codes'] = OrderedDict(self.em['codes'])
        gyro_continuous = ['GXY-GD', 'GXY-GRW']
        gyro_stationary = ['GXY-B1S', 'GXY-B2S', 'GXY-G4', 'GXY-RN']
        for tool in gyro_continuous:
            if tool in self.em['codes']:
                self.gyro_continuous = []
                self.em['codes'].move_to_end(tool)
                self.gyro_continuous.append(tool)
        self.gyro_stationary = [
            tool for tool in gyro_stationary
            if tool in self.em['codes']
        ]

        # self.em = iscwsa_error_models[model]
        #     iscwsa_error_models = yaml.safe_load(file)
        # self.em = iscwsa_error_models[model]
        if 'Default Tortusity (rad/m)' in self.em['header']:
            self.tortuosity = self.em['header']['Default Tortusity (rad/m)']
        elif 'XCL Tortuosity' in self.em['header']:
            # assuming that this is always 1 deg / 100 ft but this might not
            # be the case
            # TODO use pint to handle this string inputs
            self.tortuosity = (np.radians(1.) / 100) * 3.281
        else:
            self.tortuosity = None

        # if model == "iscwsa_mwd_rev5":
        # if model == "ISCWSA MWD Rev5":
        # assert self.tortuosity is not None, (
        #     "No default tortuosity defined in model header"
        # )

        if "Inclination Range Max" in self.em['header'].keys():
            value = np.radians(float(
                self.em['header']['Inclination Range Max'].split(" ")[0]
            ))
            assert np.amax(self.e.survey.inc_rad) < value, (
                "Model not suitable for this well path inclination"
            )

        self._initiate_func_dict()

        for err in self.em['codes']:
            # func = self._get_the_func_out(err)
            func = self.em['codes'][err]['function']
            mag = self.em['codes'][err]['magnitude']
            propagation = self.em['codes'][err]['propagation']
            self.errors[err] = (
                self.call_func(
                    code=err,
                    func=func,
                    error=self.e,
                    mag=mag,
                    propagation=propagation,
                    tortuosity=self.tortuosity,
                    header=self.em['header'],
                    errors=self
                )
            )

        self.cov_NEVs = np.zeros((3, 3, len(self.e.survey_rad)))
        for _, value in self.errors.items():
            self.cov_NEVs += value.cov_NEV

        self.cov_HLAs = NEV_to_HLA(self.e.survey_rad, self.cov_NEVs)

    def _get_the_func_out(self, err):
        if err in self.exceptional_funcs:
            func = self.exceptional_funcs[err]
        else:
            func = self.em['codes'][err]['function']

        return func

    def call_func(self, code, func, error, mag, propagation, **kwargs):
        """
        Function for calling functions by mapping function labels to their
        functions.
        """
        assert func in self.func_dict, f"no function for function {func}"

        return self.func_dict[func](code, error, mag, propagation, **kwargs)

    def _initiate_func_dict(self):
        """
        This dictionary will need to be updated if/when additional error
        functions are added to the model.
        """
        self.func_dict = {
            'ABXY_TI1': ABXY_TI1,
            'ABXY_TI2': ABXY_TI2,
            'ABZ': ABZ,
            'AMIL': AMIL,
            'ASXY_TI1': ASXY_TI1,
            'ASXY_TI2': ASXY_TI2,
            'ASXY_TI3': ASXY_TI3,
            'ASZ': ASZ,
            'DBH': DBH,
            'AZ': AZ,
            'DREF': DREF,
            'DSF': DSF,
            'DST': DST,
            'MBXY_TI1': MBXY_TI1,
            'MBXY_TI2': MBXY_TI2,
            'MBZ': MBZ,
            'MSXY_TI1': MSXY_TI1,
            'MSXY_TI2': MSXY_TI2,
            'MSXY_TI3': MSXY_TI3,
            'MSZ': MSZ,
            'SAG': SAG,
            'XYM1': XYM1,
            'XYM2': XYM2,
            'XYM3': XYM3,
            'XYM4': XYM4,
            'SAGE': SAGE,
            'XCL': XCL,  # requires an exception
            'XYM3L': XYM3L,  # looks like there's a mistake in the ISCWSA model
            'XYM4L': XYM4L,
            'XCLA': XCLA,
            'XCLH': XCLH,
            'XYM3E': XYM3E,  # Needs QAQC
            'XYM4E': XYM4E,  # Need QAQC
            'ASIXY_TI1': ASIXY_TI1,  # Needs QAQC
            'ASIXY_TI2': ASIXY_TI2,  # Needs QAQC
            'ASIXY_TI3': ASIXY_TI3,  # Needs QAQC
            'ABIXY_TI1': ABIXY_TI1,  # Needs QAQC
            'ABIXY_TI2': ABIXY_TI2,  # Needs QAQC
            'ABIZ': ABIZ,  # Needs QAQC
            'ASIZ': ASIZ,  # Needs QAQC
            'MBIXY_TI1': MBIXY_TI1,  # Needs QAQC
            'MBIXY_TI2': MBIXY_TI2,  # Needs QAQC
            'MDI': MDI,  # Needs QAQC
            'AXYZ_MIS': AXYZ_MIS,  # Needs QAQC
            'AXYZ_SF': AXYZ_SF,  # Needs QAQC
            'AXYZ_ZB': AXYZ_ZB,  # Needs QAQC
            'GXY_B1': GXY_B1,  # Needs QAQC
            'GXY_B2': GXY_B2,  # Needs QAQC
            'GXY_G1': GXY_G1,  # Needs QAQC
            'GXY_G4': GXY_G4,  # Needs QAQC
            'GXY_RN': GXY_RN,  # Needs QAQC
            'GXY_GD': GXY_GD,  # Needs QAQC
            'GXY_GRW': GXY_GRW,  # Needs QAQC
            'MFI': MFI,  # Needs QAQC
            'MSIXY_TI1': MSIXY_TI1,  # Needs QAQC
            'MSIXY_TI2': MSIXY_TI1,  # Needs QAQC
            'MSIXY_TI3': MSIXY_TI1,  # Needs QAQC
            'AMID': AMID,  # Needs QAQC
            'CNA': CNA,  # Needs QAQC
            'CNI': CNI,  # Needs QAQC
        }


def _funky_denominator(error):
    with np.errstate(divide='ignore', invalid='ignore'):
        result = np.nan_to_num((
            1 - sin(error.survey.inc_rad) ** 2
            * sin(error.survey.azi_mag_rad) ** 2
            ),
            # nan=1e-6,
            # posinf=1.0,
            # neginf=-1.0
        )
    # ACCURACY = 1e-6
    # with np.errstate(divide='ignore', invalid='ignore'):
    #     coeff = np.nan_to_num(
    #         result / np.abs(result) * ACCURACY,
    #         nan=ACCURACY
    #     )
    # result = np.where(np.abs(result) > ACCURACY, result, coeff)
    return result


# error functions #
def DREF(code, error, mag=0.35, propagation='random', NEV=True, **kwargs):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DSF(
    code, error, mag=0.00056, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    dpde = dpde * np.array(error.survey_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DST(
    code, error, mag=0.00000025, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    dpde[:, 0] = error.survey.tvd
    dpde = dpde * np.array(error.survey_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABIZ(
    code, error, mag=0.0040, propagation='systematic', NEV=True, **kwargs
):
    denom = _funky_denominator(error) / error.survey.header.G
    denom = np.where(denom > ACCURACY, denom, ACCURACY)

    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -sin(error.survey.inc_rad) / error.survey.header.G
    dpde[:, 2] = (
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad) * cos(error.survey.azi_mag_rad)
        )
    ) / denom

    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABIXY_TI1(
    code, error, mag=0.0040, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -cos(error.survey.inc_rad) / error.survey.header.G
    dpde[:, 2] = (
        cos(error.survey.inc_rad) ** 2
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad) * cos(error.survey.azi_mag_rad)
        )
    ) / (
        error.survey.header.G * (
            _funky_denominator(error)
        )
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABXY_TI1(
    code, error, mag=0.0040, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -cos(error.survey.inc_rad) / error.survey.header.G
    dpde[:, 2] = (
        cos(error.survey.inc_rad)
        * tan(error.survey.header.dip)
        * sin(error.survey.azi_mag_rad)
    ) / error.survey.header.G
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABIXY_TI2(
    code, error, mag=0.004, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            (
                -(
                    tan(error.survey.header.dip)
                    * cos(error.survey.azi_mag_rad)
                    - tan(
                        pi/2 - error.survey.inc_rad
                    )
                ) / (
                    error.survey.header.G
                    * (
                        _funky_denominator(error)
                    )
                )
            ),
            posinf=0.0,
            neginf=0.0
        )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['double_delta_md']
            * -sin(error.drdp_sing['azi2']) * mag
        ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['double_delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )

        e_NEV_sing[1, 1] = (
            (
                error.survey.md[2]
                + error.survey.md[1]
                - 2 * error.survey.md[0]
            ) / 2
            * mag * cos(error.survey.azi_true_rad[1])
            / error.survey.header.G
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['delta_md']
                * -sin(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )

        e_NEV_star_sing[1, 1] = (
            (error.survey.md[1] - error.survey.md[0])
            * mag
            * (
                cos(error.survey.azi_true_rad[1])
                / error.survey.header.G
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def ABXY_TI2(
    code, error, mag=0.004, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            (
                (
                    tan(-(error.survey_rad[:, 1]) + (pi/2))
                    - tan(error.survey.header.dip)
                    * cos(error.survey.azi_mag_rad)
                ) / error.survey.header.G
            ),
            posinf=0.0,
            neginf=0.0
        )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['double_delta_md']
            * -sin(error.drdp_sing['azi2']) * mag
        ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['double_delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        if error.error_model.lower().split(' ')[-1] != 'rev4':
            e_NEV_sing[1, 1] = (
                (
                    error.survey.md[2]
                    + error.survey.md[1]
                    - 2 * error.survey.md[0]
                ) / 2
                * mag * cos(error.survey.azi_true_rad[1])
                / error.survey.header.G
            )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['delta_md']
                * -sin(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        if error.error_model.lower().split(' ')[-1] != 'rev4':
            e_NEV_star_sing[1, 1] = (
                (error.survey.md[1] - error.survey.md[0])
                * mag
                * (
                    cos(error.survey.azi_true_rad[1])
                    / error.survey.header.G
                )
            )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def AMID(code, error, mag=0.04363323129985824, propagation='systematic',
    NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
    )
    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def ABZ(code, error, mag=0.004, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -sin(np.array(error.survey_rad)[:, 1]) / error.survey.header.G
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip) * sin(error.survey.azi_mag_rad)
    ) / error.survey.header.G
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI1(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad)
    ) / sqrt(2)
    dpde[:, 2] = (
        sin(error.survey.inc_rad)
        * -tan(error.survey.header.dip)
        * cos(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
    ) / sqrt(2)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASIXY_TI1(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad)
        / sqrt(2)
    )
    dpde[:, 2] = -(
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad) ** 2
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        )
    ) / (
        sqrt(2) * _funky_denominator(error)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI2(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(
        np.array(error.survey_rad)[:, 1]
    ) * cos(np.array(error.survey_rad)[:, 1]) / 2
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * -tan(error.survey.header.dip) * cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASIXY_TI2(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad)
        / 2
    )
    dpde[:, 2] = -(
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad) ** 2
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad) * cos(error.survey.azi_mag_rad)
        )
    ) / (
        2 * _funky_denominator(error)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI3(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip) * cos(error.survey.azi_mag_rad)
        - cos(np.array(error.survey_rad)[:, 1])) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASIXY_TI3(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        tan(error.survey.header.dip)
        * sin(error.survey.inc_rad)
        *  cos(error.survey.azi_mag_rad)
        - cos(error.survey.inc_rad)
    ) / (
        2 * _funky_denominator(error)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASZ(code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * cos(np.array(error.survey_rad)[:, 1])
    )
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip)
        * cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASIZ(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        -sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad)
    )
    dpde[:, 2] = (
        sin(error.survey.inc_rad)
        * cos(error.survey.inc_rad) ** 2
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        )
    ) / (
        _funky_denominator(error)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def AXYZ_MIS(
    code, error, mag=0.0001658062789394613, propagation='systematic', NEV=True,
    **kwargs
):
    """
    SPE 90408 Table 1
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 1., 0.])
    dpde = dpde * np.array(error.survey_rad)
    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def AXYZ_SF(
    code, error, mag=0.000111, propagation='systematic', NEV=True,
    **kwargs
):
    """
    SPE 90408 Table 1
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 1., 0.])
    dpde[:, 1] = (
        1.3 * sin(error.survey.inc_rad) * cos(error.survey.inc_rad)
    )
    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def AXYZ_ZB(
    code, error, mag=0.0017, propagation='systematic', NEV=True,
    **kwargs
):
    """
    SPE 90408 Table 1
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 1., 0.])
    dpde[:, 1] = (
        sin(error.survey.inc_rad) / error.survey.header.G
    )
    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def _get_ref_init_error(dpde, error, **kwargs):
    """
    Function that identifies where the continuous gyro begins, initiates and
    then carries the static errors during the continuous modes.
    """
    temp = [0.0]
    for coeff, inc in zip(dpde[1:, 2], error.survey.inc_rad[1:]):
        if inc > kwargs['header']['XY Static Gyro']['End Inc']:
            temp.append(temp[-1])
        else:
            temp.append(coeff)
    dpde[:, 2] = temp

    return dpde


def CNA(
    code, error, mag=0.35, propagation='systematic', NEV=True,
    **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 0.])
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            1 / sin(error.survey.inc_rad),
            posinf=1,
            neginf=-1
        )

    e_DIA = dpde * mag

    sing = np.where(
        error.survey.inc_rad < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = (
            np.array(0.5 * error.drdp_sing['double_delta_md'])
            * -sin(getattr(
                error.survey, f"azi_{error.survey.header.azi_reference}_rad"
            )[1: -1])
            * mag
        )
        e = (
            np.array(0.5 * error.drdp_sing['double_delta_md'])
            * cos(getattr(
                error.survey, f"azi_{error.survey.header.azi_reference}_rad"
            )[1: -1])
            * mag
        )
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = (
            np.array(0.5 * error.drdp_sing['delta_md'])
            * -sin(getattr(
                error.survey, f"azi_{error.survey.header.azi_reference}_rad"
            )[1: -1])
            * mag
        )
        e = (
            np.array(0.5 * error.drdp_sing['delta_md'])
            * cos(getattr(
                error.survey, f"azi_{error.survey.header.azi_reference}_rad"
            )[1: -1])
            * mag
        )
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )

    # result = error._generate_error(code, e_DIA, propagation, NEV)

    # return result


def CNI(
    code, error, mag=0.35, propagation='systematic', NEV=True,
    **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [0., 1., 0.])

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_B1(
    code, error, mag=0.002617993877991494, propagation='random',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 4
    """

    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    dpde[:, 2] = np.where(
        error.survey.inc_rad <= kwargs['header']['XY Static Gyro']['End Inc'],
        sin(error.survey.azi_true_rad)
        / (
            error.survey.header.earth_rate
            * cos(np.radians(error.survey.header.latitude))
            * cos(error.survey.inc_rad)
        ),
        np.zeros_like(error.survey.md)
    )
    dpde = _get_ref_init_error(dpde, error, **kwargs)

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_B2(
    code, error, mag=0.002617993877991494, propagation='random',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 4
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    dpde[:, 2] = np.where(
        error.survey.inc_rad <= kwargs['header']['XY Static Gyro']['End Inc'],
        cos(error.survey.azi_true_rad)
        / (
            error.survey.header.earth_rate
            * cos(np.radians(error.survey.header.latitude))
        ),
        np.zeros_like(error.survey.md)
    )
    dpde = _get_ref_init_error(dpde, error, **kwargs)

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_G1(
    code, error, mag=0.006981317007977318, propagation='systematic',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 4
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    dpde[:, 2] = np.where(
        error.survey.inc_rad <= kwargs['header']['XY Static Gyro']['End Inc'],
        cos(error.survey.azi_true_rad) * sin(error.survey.inc_rad)
        / (
            error.survey.header.earth_rate
            * cos(np.radians(error.survey.header.latitude))
        ),
        np.zeros_like(error.survey.md)
    )
    dpde = _get_ref_init_error(dpde, error, **kwargs)

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_G4(
    code, error, mag=0.010471975511965976, propagation='systematic',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 4
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    dpde[:, 2] = np.where(
        error.survey.inc_rad <= kwargs['header']['XY Static Gyro']['End Inc'],
        sin(error.survey.azi_true_rad) * tan(error.survey.inc_rad)
        / (
            error.survey.header.earth_rate
            * cos(np.radians(error.survey.header.latitude))
        ),
        np.zeros_like(error.survey.md)
    )
    dpde = _get_ref_init_error(dpde, error, **kwargs)

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_RN(
    code, error, mag=0.006981317007977318, propagation='random',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 4
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    dpde[:, 2] = np.where(
        error.survey.inc_rad <= kwargs['header']['XY Static Gyro']['End Inc'],
        1.0
        * (
            np.sqrt(
                1 - cos(error.survey.azi_true_rad) ** 2
                * sin(error.survey.inc_rad) ** 2
            )
            / (
                error.survey.header.earth_rate
                * cos(np.radians(error.survey.header.latitude))
                * cos(error.survey.inc_rad)
            )
        ),
        np.zeros_like(error.survey.md)
    )
    dpde = _get_ref_init_error(dpde, error, **kwargs)
    dpde_systematic = np.zeros_like(dpde)
    index_systematic = np.where(
        error.survey.inc_rad > kwargs['header']['XY Static Gyro']['End Inc']
    )
    np.put(
        dpde_systematic[:, 2],
        index_systematic,
        (
            dpde[index_systematic][:, 2]
            * kwargs['header']['Noise Reduction Factor']
        )
    )
    e_DIA_systematic = dpde_systematic * mag

    result_systematic = error._generate_error(
        code, e_DIA_systematic, 'systematic', NEV
    )

    np.put(
        dpde[:, 2],
        index_systematic,
        np.zeros(len(index_systematic))
    )

    # dpde[:, 2] = np.where(
    #     error.survey.inc_rad > kwargs['header']['XY Static Gyro']['End Inc'],
    #     dpde[:, 2],
    #     dpde[:, 2] * kwargs['header']['Noise Reduction Factor'],
    # )
    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    result.cov_NEV += result_systematic.cov_NEV

    return result


def GXY_GD(
    code, error, mag=0.008726646259971648, propagation='systematic',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 7
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.where(
            error.survey.inc_rad > kwargs['header']['XY Static Gyro']['End Inc'],
            np.append(
                np.array([0]),
                (
                    (error.survey.md[1:] - error.survey.md[:-1])
                    / (
                        float(
                            kwargs['header']['XY Continuous Gyro']['Running Speed'].split()[0]
                        )
                        * sin(
                            (error.survey.inc_rad[1:] + error.survey.inc_rad[:-1])
                            / 2
                        )
                    )
                )
            ),
            np.zeros_like(error.survey.md)
        )

    init_error = []
    for i, (u, l) in enumerate(zip(
        error.survey.inc_rad[1:], error.survey.inc_rad[:-1]
    )):
        init_error.append(0.0)
        if all((
            u > kwargs['header']['XY Static Gyro']['End Inc'],
            l <= kwargs['header']['XY Static Gyro']['End Inc']
        )):
            for tool in kwargs['errors'].gyro_stationary:
                temp = kwargs['errors'].errors[tool].e_DIA[i - 1][2]
                if tool in ['GXY_RN']:
                    temp *= kwargs['header']['Noise Reduction Factor']
                init_error[-1] += temp

    temp = [0.0]
    for i, (u, e) in enumerate(zip(dpde[1:, 2], init_error)):
        temp.append(0.0)
        if u != 0.0:
            temp[-1] += temp[-2] + u * mag
    dpde[:, 2] = temp

    e_DIA = dpde

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def GXY_GRW(
    code, error, mag=0.004363323129985824, propagation='systematic',
    NEV=True, **kwargs
):
    """
    SPE 90408 Table 7
    """
    dpde = np.full((len(error.survey_rad), 3), [0., 0., 1.])
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.where(
            error.survey.inc_rad > kwargs['header']['XY Static Gyro']['End Inc'],
            np.append(
                np.array([0]),
                (error.survey.md[1:] - error.survey.md[:-1])
                / (
                    float(
                        kwargs['header']['XY Continuous Gyro']['Running Speed'].split()[0]
                    )
                    * sin(
                        (error.survey.inc_rad[1:] + error.survey.inc_rad[:-1])
                        / 2
                    ) ** 2
                )
            ),
            np.zeros_like(error.survey.md)
        )

    init_error = []
    for i, (u, l) in enumerate(zip(
        error.survey.inc_rad[1:], error.survey.inc_rad[:-1]
    )):
        init_error.append(0.0)
        if all((
            u > kwargs['header']['XY Static Gyro']['End Inc'],
            l <= kwargs['header']['XY Static Gyro']['End Inc']
        )):
            for tool in kwargs['errors'].gyro_stationary:
                temp = kwargs['errors'].errors[tool].e_DIA[i - 1][2]
                if tool in ['GXY_RN']:
                    temp *= kwargs['header']['Noise Reduction Factor']
                init_error[-1] += temp

    temp = [0.0]
    for i, (u, e) in enumerate(zip(dpde[1:, 2], init_error)):
        temp.append(0.0)
        if u != 0.0:
            temp[-1] += np.sqrt(temp[-2] ** 2 + u * mag)
    dpde[:, 2] = temp

    e_DIA = dpde

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def MBXY_TI1(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / (error.survey.header.b_total * cos(error.survey.header.dip))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBIXY_TI1(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -cos(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
    ) / (
        error.survey.header.b_total
        * cos(error.survey.header.dip)
        * (
            _funky_denominator(error)
        )
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBXY_TI2(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        cos(error.survey.azi_mag_rad)
        / (
            error.survey.header.b_total
            * cos(error.survey.header.dip)
        )
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBIXY_TI2(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        cos(error.survey.azi_mag_rad)
        / (
            error.survey.header.b_total
            * cos(error.survey.header.dip)
            * (
                _funky_denominator(error)
            )
        )
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBZ(code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / (error.survey.header.b_total * cos(error.survey.header.dip))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MFI(
    code, error, mag=70, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        ) / (
            _funky_denominator(error)
        )
        / error.survey.header.b_total
    )

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def MSXY_TI1(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(np.array(error.survey_rad)[:, 1])
            + sin(np.array(error.survey_rad)[:, 1])
            * cos(error.survey.azi_mag_rad)
        ) / sqrt(2)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSXY_TI2(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(error.survey.azi_mag_rad) * (
            tan(error.survey.header.dip)
            * sin(np.array(error.survey_rad)[:, 1])
            * cos(np.array(error.survey_rad)[:, 1])
            - cos(np.array(error.survey_rad)[:, 1])
            * cos(np.array(error.survey_rad)[:, 1])
            * cos(error.survey.azi_mag_rad) - cos(error.survey.azi_mag_rad)
        ) / 2
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSXY_TI3(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        cos(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad) * cos(error.survey.azi_mag_rad)
        - cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad) * sin(error.survey.azi_mag_rad)
        - tan(error.survey.header.dip) * sin(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad)
    ) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSIXY_TI1(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(error.survey.inc_rad)
            + sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        ) / (
            sqrt(2)
            * (
                _funky_denominator(error)
            )
        )
    )

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def MSIXY_TI2(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * sin(error.survey.inc_rad)
            * cos(error.survey.inc_rad)
            - cos(error.survey.inc_rad) ** 2
            * cos(error.survey.azi_mag_rad)
            - cos(error.survey.azi_mag_rad)
        ) / (
            2 * (
                _funky_denominator(error)
            )
        )
    )

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def MSIXY_TI3(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        (
            cos(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad) ** 2
            - cos(error.survey.inc_rad)
            * sin(error.survey.azi_mag_rad) ** 2
            - tan(error.survey.header.dip)
            * sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        ) / (
            2 * (
                _funky_denominator(error)
            )
        )
    )

    e_DIA = dpde * mag

    result = error._generate_error(code, e_DIA, propagation, NEV)

    return result


def MSZ(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = -(
        sin(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad)
        + tan(error.survey.header.dip) * cos(np.array(error.survey_rad)[:, 1])
    ) * sin(np.array(error.survey_rad)[:, 1]) * sin(error.survey.azi_mag_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def AZ(code, error, mag=0.00628, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DBH(
    code, error, mag=np.radians(0.09), propagation='systematic', NEV=True,
    **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1 / (
        error.survey.header.b_total * cos(error.survey.header.dip)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MDI(
    code, error, mag=np.radians(5000), propagation='systematic', NEV=True,
    **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(error.survey.inc_rad)
        * sin(error.survey.azi_mag_rad)
        * (
            cos(error.survey.inc_rad)
            - tan(error.survey.header.dip)
            * sin(error.survey.inc_rad)
            * cos(error.survey.azi_mag_rad)
        )
    ) / (
        _funky_denominator(error)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DBHR(
    code, error, mag=np.radians(3000), propagation='random', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1 / (
        error.survey.header.b_total * cos(error.survey.header.dip)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def AMIL(code, error, mag=220.0, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
        / (error.survey.header.b_total * cos(error.survey.header.dip))
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def SAG(
    code, error, mag=0.00349, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(np.array(error.survey_rad)[:, 1])
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def SAGE(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(np.array(error.survey.inc_rad)) ** 0.25
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM1(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = np.absolute(sin(np.array(error.survey.inc_rad)))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM2(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    propagation = 'systematic'  # incorrect in the rev5 model tab
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = -1
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM3(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        np.absolute(cos(np.array(error.survey_rad)[:, 1]))
        * cos(error.survey.azi_true_rad)
    )
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            -(
                np.absolute(cos(np.array(error.survey_rad)[:, 1]))
                * sin(error.survey.azi_true_rad)
            ) / sin(np.array(error.survey_rad)[:, 1]),
            posinf=0.0,
            neginf=0.0
        )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.array(0.5 * error.drdp_sing['double_delta_md'] * mag)
        e = np.zeros(len(error.drdp_sing['double_delta_md']))
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.array(0.5 * error.drdp_sing['delta_md'] * mag)
        e = np.zeros(len(error.drdp_sing['delta_md']))
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XYM3E(code, error, mag=0.00524, propagation='random', NEV=True, **kwargs):
    coeff = np.ones(len(error.survey.md))
    coeff[1:-1] = np.amax(np.stack((
        coeff[1:-1],
        sqrt(
            10 / error.drdp_sing['delta_md']
        )
    ), axis=-1), axis=-1)
    coeff[-1] = np.amax(np.stack((
        coeff[-1],
        sqrt(
            10 / (error.survey.md[-1] - error.survey.md[-2])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey.md), 3))
    dpde[1:, 1] = np.absolute(
        cos(error.survey.inc_rad[1:])
        * cos(error.survey.azi_true_rad[1:])
        * coeff[1:]
    )

    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[1:, 2] = (
            (
                -np.absolute(cos(error.survey.inc_rad[1:]))
                * sin(error.survey.azi_true_rad[1:])
                / sin(error.survey.inc_rad[1:])
            )
            * coeff[1:]
        )
    dpde[1:, 2] = np.where(
        error.survey.inc_rad[1:] < error.survey.header.vertical_inc_limit,
        coeff[1:],
        dpde[1:, 2]
    )

    e_DIA = dpde * mag

    sing = np.where(
        error.survey.inc_rad < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[:, 0] = e_NEV[:, 0]
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV_star)
        e_NEV_star_sing[:, 0] = e_NEV_star[:, 0]
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM4(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = np.absolute(
        cos(np.array(error.survey_rad)[:, 1])
    ) * sin(error.survey.azi_true_rad)
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            (
                np.absolute(np.cos(np.array(error.survey_rad)[:, 1]))
                * cos(error.survey.azi_true_rad)
            )
            / sin(np.array(error.survey_rad)[:, 1]),
            posinf=0.0,
            neginf=0.0
            )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.zeros(len(error.drdp_sing['double_delta_md']))
        e = np.array(0.5 * error.drdp_sing['double_delta_md'] * mag)
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.zeros(len(error.drdp_sing['delta_md']))
        e = np.array(0.5 * error.drdp_sing['delta_md'] * mag)
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XYM4E(code, error, mag=0.00524, propagation='random', NEV=True, **kwargs):
    coeff = np.ones(len(error.survey.md))
    coeff[1:-1] = np.amax(np.stack((
        coeff[1:-1],
        sqrt(
            10 / error.drdp_sing['delta_md']
        )
    ), axis=-1), axis=-1)
    coeff[-1] = np.amax(np.stack((
        coeff[-1],
        sqrt(
            10 / (error.survey.md[-1] - error.survey.md[-2])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey.md), 3))
    dpde[1:, 1] = (
        cos(error.survey.inc_rad[1:])
        * sin(error.survey.azi_true_rad[1:])
        * coeff[1:]
    )

    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[1:, 2] = np.nan_to_num(
            (
                (
                    cos(error.survey.inc_rad[1:])
                    * cos(error.survey.azi_true_rad[1:])
                    / sin(error.survey.inc_rad[1:])
                )
                * coeff[1:]
            ),
            posinf=0,
            neginf=0
        )

    e_DIA = dpde * mag

    sing = np.where(
        error.survey.inc_rad < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        # this is a bit of a cop out way of handling these exceptions, but it's
        # simple and it works...
        xym3e = XYM3E(
            code, error, mag=mag, propagation=propagation, NEV=NEV
        )
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[:, 1] = xym3e.e_NEV[:, 0]
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV_star)
        e_NEV_star_sing[:, 1] = xym3e.e_NEV_star[:, 0]
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XCL(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    """
    Dummy function to manage the ISCWSA workbook not correctly defining the
    weighting functions.
    """
    tortuosity = kwargs['tortuosity']
    if code == "XCLA":
        return XCLA(
            code, error, mag=mag, propagation=propagation, NEV=NEV,
            tortuosity=tortuosity
        )
    else:
        return XCLH(
            code, error, mag=mag, propagation=propagation, NEV=NEV,
            tortuosity=tortuosity
        )


def XCLA(code, error, mag=0.167, propagation='random', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))

    def manage_sing(error, kwargs):
        temp = np.absolute(
            sin(error.survey.inc_rad[1:])
            * (((
                error.survey.azi_true_rad[1:]
                - error.survey.azi_true_rad[:-1]
                + pi
            ) % (2 * pi)) - pi)
        )
        temp[np.where(
            error.survey.inc_rad[:-1] < error.survey.header.vertical_inc_limit
        )] = 0
        return temp

    dpde[1:, 0] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            manage_sing(error, kwargs),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * -sin(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 1] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            manage_sing(error, kwargs),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.azi_true_rad[1:])
    )

    e_DIA = dpde * mag

    return error._generate_error(
        code, e_DIA, propagation, NEV, e_NEV=e_DIA, e_NEV_star=e_DIA
    )


def XCLH(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[1:, 0] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.inc_rad[1:])
        * cos(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 1] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.inc_rad[1:])
        * sin(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 2] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * -sin(error.survey.inc_rad[1:])
    )

    e_DIA = dpde * mag

    return error._generate_error(
        code, e_DIA, propagation, NEV, e_NEV=e_DIA, e_NEV_star=e_DIA
    )


def XYM3L(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    coeff = np.ones(len(error.survey.md) - 1)
    coeff = np.amax(np.stack((
        coeff,
        sqrt(
            10 / (error.survey.md[1:] - error.survey.md[:-1])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[1:, 1] = np.absolute(
        cos(error.survey.inc_rad[1:])
        * cos(error.survey.azi_true_rad[1:])
        * coeff
    )
    dpde[0, 1] = dpde[1, 1]

    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[1:, 2] = np.nan_to_num(
            (
                -np.absolute(
                    cos(error.survey.inc_rad[1:])
                )
                * (
                    sin(error.survey.azi_true_rad[1:])
                    / sin(error.survey.inc_rad[1:])
                )
                * coeff
            ),
            posinf=0,
            neginf=0
        )

    dpde[0, 2] = dpde[1, 2]

    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[1:-1, 0] = (
            coeff[:-1]
            * (
                error.survey.md[2:]
                - error.survey.md[:-2]
            ) / 2
            * mag
        )
        e_NEV_sing[1, 0] = (
            coeff[1]
            * (
                error.survey.md[2] + error.survey.md[1]
                - 2 * error.survey.md[0]
            ) / 2
            * mag
        )
        e_NEV_sing[-1, 0] = (
            coeff[-1]
            * (
                error.survey.md[-1]
                - error.survey.md[-2]
            ) / 2
            * mag
        )

        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV)
        e_NEV_star_sing[1:, 0] = (
            (
                error.survey.md[1:]
                - error.survey.md[:-1]
            ) / 2
            * mag
        )

        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XYM4L(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    propagation = 'random'
    coeff = np.ones(len(error.survey.md))
    coeff[1:] = np.amax(np.stack((
        coeff[1:],
        sqrt(
            10 / (error.survey.md[1:] - error.survey.md[:-1])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey_rad), 3))
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            np.absolute(
                cos(error.survey.inc_rad)
                * cos(error.survey.azi_true_rad)
                / sin(error.survey.inc_rad)
                * coeff
            ),
            posinf=0,
            neginf=0,
        )

    dpde[:, 1] = (
        np.absolute(
            cos(error.survey.inc_rad)
        )
        * (
            sin(error.survey.azi_true_rad)
        )
        * coeff
    )

    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[1:-1, 1] = (
            coeff[1:-1]
            * (
                error.survey.md[2:]
                - error.survey.md[:-2]
            ) / 2
            * mag
        )
        e_NEV_sing[1, 1] = (
            coeff[1]
            * (
                error.survey.md[2] + error.survey.md[1]
                - 2 * error.survey.md[0]
            ) / 2
            * mag
        )
        e_NEV_sing[-1, 1] = (
            coeff[-1]
            * (
                error.survey.md[-1]
                - error.survey.md[-2]
            ) / 2
            * mag
        )

        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV)
        e_NEV_star_sing[1:, 1] = (
            (
                error.survey.md[1:]
                - error.survey.md[:-1]
            ) / 2
            * mag
        )
        e_NEV_star_sing[1, 1] = (
            (
                error.survey.md[1]
                - error.survey.md[0]
            )
            * mag
        )

        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )



import numpy as np
from scipy.spatial.transform import Rotation as R

from numpy.typing import ArrayLike, NDArray
from typing import Any, Annotated, Literal, Union, Tuple

try:
    from numba import njit
    NUMBA = True
except ImportError:
    NUMBA = False


class MinCurve:
    def __init__(
        self,
        md,
        inc,
        azi,
        start_xyz=[0., 0., 0.],
        unit="meters"
    ):
        """
        Generate geometric data from a well bore survey.

        Parameters
        ----------
        md: list or 1d array of floats
            Measured depth along well path from a datum.
        inc: list or 1d array of floats
            Well path inclincation (relative to z/tvd axis where 0
            indicates down), in radians.
        azi: list or 1d array of floats
            Well path azimuth (relative to y/North axis),
            in radians.
        unit: str
            Either "meters" or "feet" to determine the unit of the dogleg
            severity.

        """
        assert unit == "meters" or unit == "feet", (
            'Unknown unit, please select "meters" of "feet"'
        )

        self.md = md
        survey_length = len(self.md)
        assert survey_length > 1, "Survey must have at least two rows"

        self.inc = inc
        self.azi = azi
        self.start_xyz = start_xyz
        self.unit = unit

        # make two slices with a difference or 1 index to enable array
        # calculations
        md_1 = md[:-1]
        md_2 = md[1:]

        inc_1 = inc[:-1]
        inc_2 = inc[1:]

        azi_1 = azi[:-1]
        azi_2 = azi[1:]

        # calculate the dogleg
        # temp = np.arccos(
        #     np.cos(inc_2 - inc_1)
        #     - (np.sin(inc_1) * np.sin(inc_2))
        #     * (1 - np.cos(azi_2 - azi_1))
        # )

        self.dogleg = np.zeros(survey_length)
        self.dogleg[1:] = get_dogleg(inc_1, azi_1, inc_2, azi_2)

        # calculate rf and assume rf is 1 where dogleg is 0
        self.rf = np.ones(survey_length)
        idx = np.where(self.dogleg != 0)
        self.rf[idx] = 2 / self.dogleg[idx] * np.tan(self.dogleg[idx] / 2)

        # calculate the change in md between survey stations
        temp = np.array(md_2) - np.array(md_1)
        self.delta_md = np.zeros(survey_length)
        self.delta_md[1:] = temp

        # calculate change in y direction (north)
        temp = (
            self.delta_md[1:]
            / 2
            * (
                np.sin(inc_1) * np.cos(azi_1)
                + np.sin(inc_2) * np.cos(azi_2)
            )
            * self.rf[1:]
        )
        self.delta_y = np.zeros(survey_length)
        self.delta_y[1:] = temp

        # calculate change in x direction (east)
        temp = (
            self.delta_md[1:]
            / 2
            * (
                np.sin(inc_1) * np.sin(azi_1)
                + np.sin(inc_2) * np.sin(azi_2)
            )
            * self.rf[1:]
        )
        self.delta_x = np.zeros(survey_length)
        self.delta_x[1:] = temp

        # calculate change in z direction
        temp = (
            self.delta_md[1:]
            / 2
            * (np.cos(inc_1) + np.cos(inc_2))
            * self.rf[1:]
        )
        self.delta_z = np.zeros(survey_length)
        self.delta_z[1:] = temp

        # calculate the dog leg severity
        with np.errstate(divide='ignore', invalid='ignore'):
            temp = np.degrees(self.dogleg[1:]) / self.delta_md[1:]
        self.dls = np.zeros(survey_length)
        mask = np.where(temp != np.nan)
        self.dls[1:][mask] = temp[mask]

        if unit == "meters":
            self.dls *= 30
        else:
            self.dls *= 100

        # cumulate the coordinates and add surface coordinates
        self.poss = np.vstack(
            np.cumsum(
                np.array([self.delta_x, self.delta_y, self.delta_z]).T, axis=0
            ) + self.start_xyz
        )


def get_dogleg(inc1, azi1, inc2, azi2):
    dogleg = np.arccos(
        np.cos(inc2 - inc1)
        - (np.sin(inc1) * np.sin(inc2))
        * (1 - np.cos(azi2 - azi1))
    )
    return dogleg


def get_vec(inc, azi, nev=False, r=1, deg=True):
    """
    Convert inc and azi into a vector.

    Params:
        inc: array of n floats
            Inclination relative to the z-axis (up)
        azi: array of n floats
            Azimuth relative to the y-axis
        r: float or array of n floats
            Scalar to return a scaled vector

    Returns:
        An (n,3) array of vectors
    """
    if deg:
        inc_rad, azi_rad = np.radians(np.array([inc, azi]))
    else:
        inc_rad = inc
        azi_rad = azi
    y = r * np.sin(inc_rad) * np.cos(azi_rad)
    x = r * np.sin(inc_rad) * np.sin(azi_rad)
    z = r * np.cos(inc_rad)

    if nev:
        vec = np.array([y, x, z]).T
    else:
        vec = np.array([x, y, z]).T

    return vec / np.linalg.norm(vec, axis=-1).reshape(-1, 1)


def get_nev(
    pos, start_xyz=np.array([0., 0., 0.]), start_nev=np.array([0., 0., 0.])
):
    """
    Convert [x, y, z] coordinates to [n, e, tvd] coordinates.

    Params:
        pos: (n,3) array of floats
            Array of [x, y, z] coordinates
        start_xyz: (,3) array of floats
            The datum of the [x, y, z] cooardinates
        start_nev: (,3) array of floats
            The datum of the [n, e, tvd] coordinates

    Returns:
        An (n,3) array of [n, e, tvd] coordinates.
    """
    # e, n, v = (
    #     np.array([pos]).reshape(-1,3) - np.array([start_xyz])
    # ).T
    e, n, v = (
        np.array([pos]).reshape(-1, 3) - np.array([start_xyz])
    ).T

    return (np.array([n, e, v]).T + np.array([start_nev]))


def get_xyz(pos, start_xyz=[0., 0., 0.], start_nev=[0., 0., 0.]):
    y, x, z = (
        np.array([pos]).reshape(-1, 3) - np.array([start_nev])
    ).T

    return (np.array([x, y, z]).T + np.array([start_xyz]))


def _get_angles(vec):
    xy = vec[:, 0] ** 2 + vec[:, 1] ** 2
    inc = np.arctan2(np.sqrt(xy), vec[:, 2])  # for elevation angle defined from Z-axis down
    azi = (np.arctan2(vec[:, 0], vec[:, 1]) + (2 * np.pi)) % (2 * np.pi)

    return np.stack((inc, azi), axis=1)


if NUMBA:
    _get_angles = njit(_get_angles)


def get_angles(
    vec: Annotated[NDArray, Literal["N", 3]], nev: bool = False
):
    '''
    Determines the inclination and azimuth from a vector.

    Parameters
    ----------
    vec: (n,3) array of floats
    nev: boolean (default: False)
        Indicates if the vector is in (x,y,z) or (n,e,v) coordinates.

    Returns
    -------
    [inc, azi]: (n,2) array of floats
        A numpy array of incs and axis in radians

    '''
    # make sure it's a unit vector
    vec = vec / np.linalg.norm(vec, axis=-1).reshape(-1, 1)
    vec = vec.reshape(-1, 3)

    # if it's nev then need to do the shuffle
    if nev:
        y, x, z = vec.T
        vec = np.array([x, y, z]).T

    return _get_angles(vec)


def _get_transform(inc, azi):
    trans = np.array([
        [np.cos(inc) * np.cos(azi), -np.sin(azi), np.sin(inc) * np.cos(azi)],
        [np.cos(inc) * np.sin(azi), np.cos(azi), np.sin(inc) * np.sin(azi)],
        [-np.sin(inc), np.zeros_like(inc), np.cos(inc)]
    ]).T

    return trans


def get_transform(
    survey
):
    """
    Determine the transform for transforming between NEV and HLA coordinate
    systems.

    Params:
        survey: (n,3) array of floats
        The [md, inc, azi] survey listing array.

    Returns:
        transform: (n,3,3) array of floats
    """
    survey = survey.reshape(-1, 3)
    inc = np.array(survey[:, 1])
    azi = np.array(survey[:, 2])

    return _get_transform(inc, azi)

    # trans = np.array([
    #     [np.cos(inc) * np.cos(azi), -np.sin(azi), np.sin(inc) * np.cos(azi)],
    #     [np.cos(inc) * np.sin(azi), np.cos(azi), np.sin(inc) * np.sin(azi)],
    #     [-np.sin(inc), np.zeros_like(inc), np.cos(inc)]
    # ]).T

    # return trans


def NEV_to_HLA(
    survey: Annotated[NDArray, Literal["N", 3]],
    NEV: Union[
        Annotated[NDArray, Literal["N", 3]],
        Annotated[NDArray, Literal[3, 3, "N"]]
    ],
    cov: bool = True
) -> Union[
        Annotated[NDArray, Literal['..., 3']],
        Annotated[NDArray, Literal['3, 3, ...']]
]:
    """
    Transform from NEV to HLA coordinate system.

    Parameters:
    -----------
    survey: (n,3) array of floats
        The [md, inc, azi] survey listing array.
    NEV: (d,3) or (3,3,d) array of floats
        The NEV coordinates or covariance matrices.
    cov: boolean
        If cov is True then a (3,3,d) array of covariance matrices
        is expected, else a (d,3) array of coordinates.

    Returns:
    --------
    HLAs: NDArray
        Either a transformed (n,3) array of HLA coordinates or an
        (3,3,n) array of HLA covariance matrices.
    """

    trans = get_transform(survey)

    if cov:
        # HLAs = [
        #     np.dot(np.dot(t, NEV.T[i]), t.T) for i, t in enumerate(trans)
        # ]
        # HLAs = np.vstack(HLAs).reshape(-1, 3, 3).T

        HLAs = np.einsum(
            '...ik,...jk',
            np.einsum(
                '...ik,...jk', trans, NEV.T
            ),
            trans
        ).T

    else:
        NEV = NEV.reshape(-1, 3)
        # HLAs = [
        #     np.dot(NEV[i], t.T) for i, t in enumerate(trans)
        # ]

        HLAs = np.einsum(
            '...k,...jk', NEV, trans
        )

    return HLAs


def HLA_to_NEV(survey, HLA, cov=True, trans=None):
    if trans is None:
        trans = get_transform(survey)

    if cov:
        # NEVs = [
        #     np.dot(np.dot(t.T, HLA.T[i]), t) for i, t in enumerate(trans)
        # ]
        # NEVs = np.vstack(NEVs).reshape(-1, 3, 3).T

        NEVs = np.einsum(
            '...ik,jk...',
            np.einsum(
                '...ki,...jk', trans, HLA.T
            ),
            trans.T
        ).T

    else:
        # NEVs = [
        #     np.dot(hla, t) for hla, t in zip(HLA, trans)
        # ]

        NEVs = np.einsum(
            'k...,jk...', HLA.T, trans.T
        )

    # return np.vstack(NEVs).reshape(HLA.shape)

    return np.swapaxes(NEVs, 0, 1)


def get_sigmas(cov, long=False):
    """
    Extracts the sigma values of a covariance matrix along the principle axii.

    Parameters
    ----------
        cov: (n,3,3) array of floats

    Returns
    -------
        arr: (n,3) array of floats
    """

    assert cov.shape[-2:] == (3, 3), "Cov is the wrong shape"

    cov = cov.reshape(-1, 3, 3)

    aa, ab, ac = cov[:, :, 0].T
    ab, bb, bc = cov[:, :, 1].T
    ac, bc, cc = cov[:, :, 2].T

    if long:
        return (aa, bb, cc, ab, ac, bc)
    else:
        return (np.sqrt(aa), np.sqrt(bb), np.sqrt(cc))

    # a, b, c = np.array([
    #     np.sqrt(cov[:, 0, 0]),
    #     np.sqrt(cov[:, 1, 1]),
    #     np.sqrt(cov[:, 2, 2])
    # ])

    # return (a, b, c)


def get_unit_vec(vec):
    vec = vec / np.linalg.norm(vec)

    return vec


def linear_convert(data, factor):
    flag = False
    if type(data) != list:
        flag = True
        data = [data]
    converted = [d * factor if d is not None else None for d in data]
    if flag:
        return converted[0]
    else:
        return converted


def make_cov(a, b, c, long=False):
    # a, b, c = np.sqrt(np.array([a, b, c]))
    if long:
        cov = np.array([
            [a * a, a * b, a * c],
            [a * b, b * b, b * c],
            [a * c, b * c, c * c]
        ])

    else:
        cov = np.array([
            [a * a, np.zeros_like(a), np.zeros_like(a)],
            [np.zeros_like(a), b * b, np.zeros_like(a)],
            [np.zeros_like(a), np.zeros_like(a), c * c]
        ])

    return cov.T


def dls_from_radius(radius):
    """
    Returns the dls in degrees from a radius.
    """
    if isinstance(radius, np.ndarray):
        circumference = np.full_like(radius, np.inf)
        circumference = np.where(
            radius != 0,
            2 * np.pi * radius,
            circumference
        )
    else:
        if radius == 0:
            return np.inf
        circumference = 2 * np.pi * radius
    dls = 360 / circumference * 30

    return dls


def radius_from_dls(dls):
    """
    Returns the radius in meters from a DLS in deg/30m.
    """
    if isinstance(dls, np.ndarray):
        circumference = np.full_like(dls, np.inf)
        circumference = np.where(
            dls != 0,
            (30 / dls) * 360,
            circumference
        )
    else:
        if dls == 0:
            return np.inf
        circumference = (30 / dls) * 360
    radius = circumference / (2 * np.pi)

    return radius


def errors_from_cov(cov, data=False):
    """
    Parameters
    ----------
    cov: (n, 3, 3) array
        The error covariance matrices.
    data: bool (default: False)
        If True returns a dictionary, else returns a list.
    """
    nn, ne, nv, _, ee, ev, _, _, vv = (
        cov.reshape(-1, 9).T
    )

    if data:
        return {
            i: {
                'nn': _nn, 'ne': _ne, 'nv': _nv, 'ee': _ee, 'ev': _ev, 'vv': _vv
            }
            for i, (_nn, _ne, _nv, _ee, _ev, _vv)
            in enumerate(zip(nn, ne, nv, ee, ev, vv))
        }

    return np.array([nn, ne, nv, ee, ev, vv]).T


class Arc:
    def __init__(self, dogleg, radius):
        """
        Generates a generic arc that can be transformed with a specific pos
        and vec via a transform method. The arc is initialized at a local
        origin and kicks off down and to the north (assuming an NEV coordinate
        system).

        Parameters
        ----------
        dogleg: float
            The sweep angle of the arc in radians.
        radius: float
            The radius of the arc in meters.

        Returns
        -------
        arc: Arc object
        """
        self.dogleg = dogleg
        self.radius = radius
        self.delta_md = dogleg * radius

        self.pos = np.array([
            np.cos(dogleg),
            0.,
            np.sin(dogleg)
        ]) * radius
        self.pos[0] = radius - self.pos[0]

        self.vec = np.array([
            np.sin(dogleg),
            0.,
            np.cos(dogleg)
        ])

    def transform(self, toolface, pos=None, vec=None, target=False):
        """
        Transforms an Arc to a position and orientation.

        Parameters
        ----------
        pos: (,3) array
        The desired position to transform the arc.
        vec: (,3) array
            The orientation unit vector to transform the arc.
        target: bool
            If true, returned arc vector is reversed.

        Returns
        -------
        tuple (pos_new, vec_new)
        pos_new: (,3) array
            The position at the end of the arc post transform.
        vec_new: (,3) array
            The unit vector at the end of the arc post transform.
        """
        if vec is None:
            vec = np.array([0., 0., 1.])
        if target:
            vec *= -1
        inc, azi = get_angles(vec, nev=True).reshape(2)
        angles = [
            toolface,
            inc,
            azi
        ]
        r = R.from_euler('zyz', angles, degrees=False)

        pos_new, vec_new = r.apply(np.vstack((self.pos, self.vec)))

        if pos is not None:
            pos_new += pos
        if target:
            vec_new *= -1

        return (pos_new, vec_new)


def get_arc(dogleg, radius, toolface, pos=None, vec=None, target=False):
    """
    Creates an Arc instance and transforms it to the desired position and
    orientation.

    Parameters
    ----------
    dogleg: float
        The swept angle of the arc (arc angle) in radians.
    radius: float
        The radius of the arc (in meters).
    toolface: float
        The toolface angle in radians (relative to the high side) to rotate the
        arc at the desired position and orientation.
    pos: (,3) array
        The desired position to transform the arc.
    vec: (,3) array
        The orientation unit vector to transform the arc.
    target: bool
        If true, returned arc vector is reversed.

    Returns
    -------
    tuple of (pos_new, vec_new, arc.delta_md)
    pos_new: (,3) array
        The position at the end of the arc post transform.
    vec_new: (,3) array
        The unit vector at the end of the arc post transform.
    arc.delta_md: int
        The arc length of the arc.
    """
    arc = Arc(dogleg, radius)
    pos_new, vec_new = arc.transform(toolface, pos, vec, target)

    return (pos_new, vec_new, arc.delta_md)



"""
The different setup versions method is shamelessly copied from mikedh's trimesh
Python library - thank you!
"""

import os
import sys
from setuptools import setup, find_packages

# load __version__ without importing anything
version_file = os.path.join(
    os.path.dirname(__file__),
    'welleng/version.py')
with open(version_file, 'r') as f:
    # use eval to get a clean string of version from file
    __version__ = eval(f.read().strip().split('=')[-1])

with open("README.md", "r") as f:
    long_description = f.read()

# with open("requirements.txt") as f:
#     required = f.read().splitlines()

download_url = f'https://github.com/jonnymaserati/welleng/archive/v{__version__}.tar.gz'

# If you only want to generate surveys and errors, these are all that's
# required
requirements_default = set([
    'numpy',
    'scipy',
    'pint',
    'PyYAML',
    'setuptools',
    'vtk'
])

# these can be installed without compiling required
requirements_easy = set([
    'magnetic_field_calculator',    # used to get default mag data for survey
    'networkx',
    'openpyxl',
    'pandas',
    'tabulate',
    'trimesh',
    'utm',
    'vedo',
])

# this is the troublesome requirement that needs C dependencies
requirements_all = requirements_easy.union([
    'python-fcl',
])

# if someone wants to output a requirements file
# `python setup.py --list-all > requirements.txt`
if '--list-all' in sys.argv:
    # will not include default requirements (numpy)
    print('\n'.join(requirements_all))
    exit()
elif '--list-easy' in sys.argv:
    # again will not include numpy+setuptools
    print('\n'.join(requirements_easy))
    exit()

# if sys.platform == 'win32':
#     requirements_all.append('python-fcl-win32')
# else:
#     requirements_all.append('python-fcl')

setup(
    name='welleng',
    version=__version__,
    description='A collection of Well Engineering tools',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/jonnymaserati/welleng',
    download_url=download_url,
    keywords=[
        'well',
        'trajectory',
        'wellpath',
        'wellbore',
        'drilling',
        'drill',
        'error',
        'separation',
        'minimum curvature',
        'iscwsa',
        'owsg',
        'well engineering',
        'wells',
        'drilling engineering',
        'directional drilling',
        'mwd',
        'survey',
        'covariance',
        'digitalization',
        'automation',
        'volve',
        'witsml',
    ],
    author='Jonathan Corcutt',
    author_email='jonnycorcutt@gmail.com',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.10',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    python_requires='>=3.7.*',
    packages=find_packages(exclude=["tests"]),
    package_data={
        'welleng': [
            'errors/*.yaml',
            'errors/tool_codes/*.yaml',
            'exchange/*.yaml'
        ]
    },
    install_requires=list(requirements_default),
    extras_require={
        'easy': list(requirements_easy),
        'all': list(requirements_all)
    }
)



from openpyxl import load_workbook
from math import radians
import pandas as pd
import yaml
import os

"""
Code to extract error model data from standard ISCWSA Excel files and create
yaml files for use with the welleng.error module.

It's coded up this way so that when the error models are updated then simply
add the new workbook to the reference folder and add a new line to remake
the error model input data.

If new error functions are added then these will need to be coded into the
error module.
"""

PATH = os.path.dirname(__file__)
# PATH = os.path.join('', *[PATH, 'tool_codes'])

# FILENAME = 'welleng/errors/error_codes.yaml'
CHARACTERS = [":", "[", "]"]

# try:
#     with open(FILENAME, 'r') as file:
#         ec = yaml.full_load(file)
# except Exception:
#     ec = {}


def extract_data(file, sheet_name='Model'):
    # check if file is workbook otherwise treat as filename
    try:
        m = file[sheet_name]
    except TypeError:
        workbook = load_workbook(
            filename=file,
            data_only=True
        )
        m = workbook[sheet_name]

    e = extract_codes(m)
    h = extract_header(m)

    ed = dict(
        header=h,
        codes=e
    )

    return ed


def extract_header(worksheet):
    m = worksheet
    d = []
    tag = False
    temp = {}

    for row in m.iter_rows(
        min_row=0,
        max_row=(),
        min_col=0,
        max_col=2,
        values_only=True
    ):
        if tag:
            if row[0][0] == " ":
                if 'deg' in row[1]:
                    value = radians(float(row[1].split(" ")[0]))
                else:
                    value = row[1]
                temp[tag][row[0].lstrip().replace(":", "")] = value
                continue
            else:
                tag = False
        if row[0] is None:
            continue
        if "gyro" in row[0].lower():
            tag = row[0].replace(":", "")
            temp[tag] = {}
        else:
            d.append([
                row[0], row[1]
            ])

    h = make_header_dict(d)

    if bool(temp):
        h.update(temp)

    return h


def extract_codes(worksheet):
    m = worksheet
    d = []

    for row in m.iter_rows(
        min_row=4,
        max_row=(),
        min_col=4,
        max_col=(),
        values_only=True
    ):
        code = row[1]
        if code is None:
            continue
        if code == "XCLL":
            code = "XCLA"
        function = row[3]
        magnitude = row[6]
        unit = row[7]
        propagation = row[8]
        d.append([
            code, function, magnitude, unit, propagation
        ])

    e = make_error_dict(d)

    return e


def make_error_dict(data):
    e = {}

    for c, f, m, u, p in data:
        e[c] = dict(
            function=f.replace('-', '_'),
            magnitude=m if 'deg' not in u else radians(m),
            unit=u.replace("deg", "rad"),
            propagation='random' if p == 'R' else 'systematic'
        )

    return e


def make_header_dict(data):
    h = {}

    for k, v in data:
        kk = remove_characters(k)
        h[kk] = v

    return h


def remove_characters(data, chars=CHARACTERS):
    for c in chars:
        data = data.replace(c, "")

    return data


def open_workbook(filename):
    workbook = load_workbook(
        filename=filename,
        data_only=True
    )

    return workbook


def make_index(wb):
    d = {}

    for i, row in enumerate(wb['Index'].iter_rows(
        values_only=True
    )):
        if i == 0:
            fields = row
            continue
        if row[2].split('_')[-1] == 'Fl':
            owsg_prefix = f"{row[1]}_Fl"
        else:
            owsg_prefix = row[1]
        d[owsg_prefix] = {}
        for f, r in zip(fields, row):
            d[owsg_prefix][f] = r

    return d


def get_index(short_name, index_file):
    for k, v in index_file.items():
        if v['Short Name'] == short_name:
            return k

    return None


def get_short_names(index_file):
    try:
        with open(index_file, 'r') as f:
            index_file = yaml.safe_load(f)
    except TypeError:
        pass

    short_names = {
        k: v['Short Name']
        for k, v in index_file.items()
    }

    return short_names


if __name__ == '__main__':
    sheet_ignore_list = ['Sheet']
    filename = (
        f'reference/'
        f'toolgroup-owsg-a-rev-5-1-08-oct-2020-produced-22-oct-2020.xlsx'
    )

    wb = open_workbook(
        filename
    )

    temp = make_index(wb)
    filename = os.path.join(
        '',
        *[PATH, 'tool_index.yaml']
    )

    # open tool index file
    with open(filename, 'r') as f:
        tool_index = yaml.safe_load(f)

    # add the new tools
    for k, v in temp.items():
        tool_index[k] = v

    # write the updated file
    with open(filename, 'w') as f:
        yaml.dump(tool_index, f)

    short_names = get_short_names(tool_index)

    tool = get_index([*short_names.values()][10], tool_index)

    ec = {}

    # sheets are named by 'Short Name'
    for k, v in temp.items():
        sheet_name = v['Short Name']
        for s in sheet_ignore_list:
            if s in sheet_name:
                continue

        ec[k] = extract_data(
            file=wb, sheet_name=sheet_name
        )

    for k, v in ec.items():
        owsg_prefix = v['header']['OWSG Prefix']
        filename = os.path.join(
            '',
            *[PATH, 'tool_codes', f"{k}.yaml"]
        )
        with open(filename, 'w') as f:
            yaml.dump(ec[k], f)

    # wb = open_workbook(
    #     'reference/toolgroup-owsg-a-rev-5-1-08-oct-2020-produced-22-oct-2020.xlsx'
    # )



    # ec['iscwsa_mwd_rev4'] = extract_data(
    #     "toolgroup-owsg-a-rev-5-1-08-oct-2020-produced-22-oct-2020.xlsx"
    # )

    # ec['iscwsa_mwd_rev5'] = extract_data(
    #     "reference/error-model-example-mwdrev5-iscwsa-1.xlsx"
    # )

    # with open(
    #     os.path.join("", *[PATH, "tool_codes", "ISCWSA MWD Rev5.yaml"]),
    #     'w'
    # ) as f:
    #     documents = yaml.dump(ec, f)

    print("Done")



from copy import deepcopy
import numpy as np

try:
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go
    PLOTLY = True
except ImportError:
    PLOTLY = False

from .survey import interpolate_md, Survey
from .units import ureg


class TorqueDrag:
    def __init__(
        self, survey, wellbore, string, fluid_density, name=None,
        wob=None, tob=None, overpull=None,
    ):
        """
        A class for calculating wellbore torque and drag, based on the
        "Torque and Drag in Directional Wells--Prediction and Measurement
        (SPE 11380-PA) by C.A. Johancsik et al.

        Parameters
        ----------
        survey: welleng.survey.Survey instance
            The well trajectory of the scenario being modelled.
        wellbore: welleng.architecture.WellBore instance
            The well bore architecture of the scenario being modelled.
        string: welleng.architecture.BHA or welleng.architecture.CasingString
        instance
            The string being run inside the well bore for the scenario being
            modelled.
        fluid_density: float
            The density (in SG) of the fluid in the well bore.
        name: str
            The name of the scenario being modeled.
        wob: float
            The compressive force (weight on bit) applied at the bottom of the
            string in N.
        tob: float
            The torque (torque on bit) applied at the bottom of the string in
            N.m.
        overpull: float
            The tension applied at the bottom of the string in N.
        """
        assert wellbore.complete, "Wellbore not complete"
        assert string.complete, "String not complete"

        self.survey_original = survey
        self.wellbore = wellbore
        self.string = string
        self.fluid_density = fluid_density
        self.name = name
        self.add_survey_points_from_strings()
        self.index = np.where(self.survey.md == self.string.bottom)[0][0] + 1

        self.get_buoyancy_factors()
        self.get_inc_average()
        self.get_inc_delta()
        self.get_azi_delta()
        self.get_weight_buoyed_and_radius()

        self.torque, self.tension = {}, {}
        self.get_coeff_friction_sliding()
        self.get_forces_and_torsion()
        if any((wob, tob, overpull)):
            self.get_forces_and_torsion(wob=wob, tob=tob, overpull=overpull)

    def add_survey_points_from_strings(self):
        """
        Check that there's survey stations for the top and bottoms of
        the string sections to ensure that the torque and drag is
        calculated for these key locations.
        """
        q = []

        for k, v in self.wellbore.sections.items():
            if v['bottom'] in self.survey_original.md:
                continue
            else:
                q.append(interpolate_md(
                    self.survey_original, v['bottom']
                ))

        for k, v in self.string.sections.items():
            if v['bottom'] in self.survey_original.md:
                continue
            else:
                q.append(interpolate_md(
                    self.survey_original, v['bottom']
                ))

        md, inc, azi = (
            self.survey_original.md.tolist(),
            self.survey_original.inc_rad.tolist(),
            self.survey_original.azi_grid_rad.tolist()
        )

        md.extend([station.md[-1] for station in q])
        inc.extend([station.inc_rad[-1] for station in q])
        azi.extend([station.azi_grid_rad[-1] for station in q])

        md, inc, azi = zip(*sorted(zip(md, inc, azi)))

        sh = self.survey_original.header
        sh.azi_reference = 'grid'

        self.survey = Survey(
            md=md, inc=inc, azi=azi, header=sh, deg=False
        )

    def get_buoyancy_factors(self):
        """
        Determine the buoyancy factor for each string section and add it
        to the string sections dict.
        """
        for k, v in self.string.sections.items():
            v['buoyancy_factor'] = buoyancy_factor(
                self.fluid_density, v['density']
            )

    def get_inc_average(self):
        self.inc_average = np.zeros_like(self.survey.inc_rad)
        self.inc_average[1:] = np.average(
            (self.survey.inc_rad[1:], self.survey.inc_rad[:-1]),
            axis=0
        )

    def get_inc_delta(self):
        self.inc_delta = np.zeros_like(self.survey.inc_rad)
        self.inc_delta[1:] = self.survey.inc_rad[1:] - self.survey.inc_rad[:-1]

    def get_azi_delta(self):
        self.azi_delta = np.zeros_like(self.survey.azi_grid_rad)
        self.azi_delta[1:] = (
            self.survey.azi_grid_rad[1:] - self.survey.azi_grid_rad[:-1]
        )

    def get_characteristic_od(self, section):
        if bool(self.string.sections[section].get('tooljoint_od')):
            return self.string.sections[section]['tooljoint_od']
        else:
            return self.string.sections[section]['od']

    def get_weight_buoyed_and_radius(self):
        section = 0
        weights = [0]
        diameter = [self.get_characteristic_od(section)]
        for md, delta_md in zip(self.survey.md[1:], self.survey.delta_md[1:]):
            if md > self.string.bottom:
                break
            while 1:
                if (
                    self.string.sections[section]['top']
                    < md <= self.string.sections[section]['bottom']
                ):
                    weights.append(
                        self.string.sections[section]['unit_weight'] * delta_md
                        * self.string.sections[section]['buoyancy_factor']
                    )
                    diameter.append(
                        self.get_characteristic_od(section)
                    )
                    break
                else:
                    section += 1
        self.weight_buoyed = np.array(weights)
        self.radius = np.array(diameter) / 2

    def get_coeff_friction_sliding(self):
        section = 0
        friction = [self.wellbore.sections[section]['coeff_friction_sliding']]
        for md, delta_md in zip(self.survey.md[1:], self.survey.delta_md[1:]):
            if md > self.wellbore.bottom:
                break
            while 1:
                if (
                    self.wellbore.sections[section]['top']
                    < md <= self.wellbore.sections[section]['bottom']
                ):
                    friction.append(
                        self.wellbore.sections[section]['coeff_friction_sliding']  # noqa: E501
                    )
                    break
                else:
                    section += 1
        self.coeff_friction_sliding = np.array(friction)

    def get_forces_and_torsion(self, wob=False, tob=False, overpull=False):
        if any((wob, tob)):
            assert tob, "Can't have WOB without TOB"
            assert wob, "Can't have TOB wihtouh WOB"
            ft = [np.array([0.0, -wob, -wob])]
            tn = [tob]
        else:
            ft = [np.zeros(3)]
            tn = [0]
        if overpull:
            ft[0][0] = overpull
        fn = []
        for row in zip(
            self.inc_average[:self.index][::-1],
            self.inc_delta[:self.index][::-1],
            self.azi_delta[:self.index][::-1],
            self.weight_buoyed[::-1],
            self.coeff_friction_sliding[:self.index][::-1],
            self.radius
        ):
            (
                inc_average, inc_delta, azi_delta, weight_buoyed,
                coeff_friction_sliding, radius
            ) = row

            fn.append(force_normal(
                ft[-1], inc_average, inc_delta, azi_delta,
                weight_buoyed
            ))
            ft.append(ft[-1] + np.array(force_tension_delta(
                weight_buoyed, inc_average, coeff_friction_sliding, fn[-1]
            )))
            tn.append(tn[-1] + np.array(torsion_delta(
                coeff_friction_sliding, fn[-1][2], radius
            )))

        fn = np.array(fn)[::-1]
        ft = np.array(ft)[::-1][1:]
        tn = np.array(tn)[::-1][1:]
        if wob:
            self.tension["drilling"] = ft[:, 2]
            self.tension["sliding"] = ft[:, 1]
        else:
            self.tension['slackoff'] = ft[:, 1]
            self.tension['rotating'] = ft[:, 2]

        if tob:
            self.torque['drilling'] = tn
        else:
            self.torque['rotating'] = tn

        if overpull:
            self.tension['overpull'] = ft[:, 0]
        else:
            self.tension['pickup'] = ft[:, 0]

        self.wob, self.tob, self.overpull = wob, tob, overpull

    def figure(self):
        return figure_string_tension_and_torque(self)


def force_normal(
    force_tension,
    inc_average,
    inc_delta,
    azi_delta,
    weight_buoyed,
):
    result = np.sqrt(
        (force_tension * azi_delta * np.sin(inc_average)) ** 2
        + (
            force_tension * inc_delta
            + weight_buoyed * np.sin(inc_average)
        ) ** 2
    )

    return result


def force_tension_delta(
    weight_buoyed,
    inc_average,
    coeff_friction_sliding,
    force_normal
):
    A = weight_buoyed * np.cos(inc_average)
    B = coeff_friction_sliding * force_normal

    pickup, slackoff, rotating = A + B * np.array([1, -1, 0])

    return (pickup, slackoff, rotating)


def torsion_delta(
    coeff_friction_sliding,
    force_normal,
    radius
):
    result = coeff_friction_sliding * force_normal * radius

    return result


def buoyancy_factor(fluid_density, string_density=7.85):
    """
    Parameters
    ----------
    fluid_density: float
        The density of the fluid in SG.
    string_density: float
        The density of the string, typically made from steel.

    Returns
    -------
    result: float
        The buoyancy factor when when multiplied against the string weight
        yields the bouyed string weight.
    """
    result = (string_density - fluid_density) / string_density

    return result


class HookLoad:
    def __init__(
        self, survey, wellbore, string, fluid_density, step=30, name=None,
        ff_range=(0.1, 0.4, 0.1)
    ):
        """
        A class for calculating the hookload or broomstick plot data for
        running or pulling a string in a wellbore.

        Parameters
        ----------
        survey: welleng.survey.Survey instance
            The well trajectory of the scenario being modelled.
        wellbore: welleng.architecture.WellBore instance
            The well bore architecture of the scenario being modelled.
        string: welleng.architecture.BHA or welleng.architecture.CasingString
        instance
            The string being run inside the well bore for the scenario being
            modelled.
        fluid_density: float
            The density (in SG) of the fluid in the well bore.
        step: float
            The measured depth step distance in meters to move the string.
        name: str
            The name of the scenario being modeled.
        ff_range: (3) tuple
            The start, stop and step for the range of friction factors to be
            used in the hookload calculations.
        """
        self.survey = survey
        self.wellbore = wellbore
        self.string = string
        self.fluid_density = fluid_density
        self.name = name
        self.step = step
        self.get_ff_range(ff_range)
        self.get_data()

    def get_ff_range(self, ff_range):
        self.ff_range = np.arange(*ff_range).tolist()
        self.ff_range.append(ff_range[1])

    def get_data(self):
        self.data = {}
        self.md_range = np.arange(
            self.string.top + self.step, self.string.bottom, self.step
        ).tolist()
        self.md_range.append(self.string.bottom)

        for ff in self.ff_range:
            wellbore_temp = deepcopy(self.wellbore)
            for k in wellbore_temp.sections.keys():
                wellbore_temp.sections[k]['coeff_friction_sliding'] = ff

            self.data[ff] = {}
            data_temp = []

            for md in self.md_range:
                bha_temp = self.string.depth(md)
                data_temp.append(
                    TorqueDrag(
                        self.survey, wellbore_temp, bha_temp,
                        fluid_density=self.fluid_density,
                        name=self.name
                    )
                )

            for t in data_temp[0].tension.keys():
                self.data[ff][t] = [
                    d.tension[t][0] for d in data_temp
                ]

    def figure(self):
        return figure_hookload(self)


def figure_string_tension_and_torque(
    td,
    units=dict(
        depth='ft',
        tension='lbf',
        torque='ft_lbf'
    )
):
    assert PLOTLY, "Please install plotly"

    fig = make_subplots(rows=1, cols=2)

    for k, v in td.tension.items():
        fig.add_trace(
            go.Scatter(
                x=((v * ureg('N')).to(units['tension'])).m,
                y=((td.survey.md * ureg.meters).to(units['depth'])).m,
                name=f"tension: {k}",
            ),
            row=1, col=1
        )

    for k, v, in td.torque.items():
        fig.add_trace(
            go.Scatter(
                x=((v * ureg('Nm')).to(units['torque'])).m,
                y=((td.survey.md * ureg.meters).to(units['depth'])).m,
                name=f"torque: {k}",
            ),
            row=1, col=2
        )

    fig.update_layout(
        title_text=(
            f"<b>wellbore:</b> {td.wellbore.name}<br>"
            + f"<b>string:</b> {td.string.name}"
        ),
        xaxis=dict(
            title=f"Tension ({units['tension']})"
        ),
        yaxis=dict(
            autorange='reversed',
            title=f"MD ({units['depth']})",
            exponentformat='none',
        ),
        xaxis2=dict(
            title=f"Torque ({units['torque']})"
        ),
        yaxis2=dict(
            autorange='reversed',
            title=f"MD ({units['depth']})",
            exponentformat='none',
        )
    )

    return fig


def figure_hookload(
    hl,
    units=dict(
        depth='ft',
        tension='lbf',
        torque='ft_lbf'
    )
):
    assert PLOTLY, "Please install plotly"

    fig = go.Figure()

    lines = [None, 'dashdot', 'dash', 'dot']

    md = ((hl.md_range * ureg.meters).to(units['depth'])).m
    annotations = []

    for i, (k, v) in enumerate(hl.data.items()):
        x = ((v['slackoff'] * ureg('N')).to(units['tension'])).m
        fig.add_trace(
            go.Scatter(
                x=x,
                y=md,
                name=f"SOFF: {k:.2f}",
                line=dict(color='blue', dash=lines[i])
            ),
        )
        annotations.append(
            dict(
                x=x[-1], y=md[-1],
                xanchor='center', yanchor='top',
                text=f'{k:.2f}',
                showarrow=False,
                font=dict(
                    color='blue'
                )
            )
        )

    for i, (k, v) in enumerate(hl.data.items()):
        fig.add_trace(
            go.Scatter(
                x=((v['rotating'] * ureg('N')).to(units['tension'])).m,
                y=md,
                name=f"RoffBFF: {k:.2f}",
                line=dict(color='green')
            )
        )
        break

    for i, (k, v) in enumerate(hl.data.items()):
        x = ((v['pickup'] * ureg('N')).to(units['tension'])).m
        fig.add_trace(
            go.Scatter(
                x=x,
                y=md,
                name=f"PUFF: {k:.2f}",
                line=dict(color='red', dash=lines[i])
            ),
        )
        annotations.append(
            dict(
                x=x[-1], y=md[-1],
                xanchor='center', yanchor='top',
                text=f'{k:.2f}',
                showarrow=False,
                font=dict(
                    color='red'
                )
            )
        )

    title_text = ("<b>wellbore:</b>")
    for i, (k, v) in enumerate(hl.wellbore.sections.items()):
        coupler = "" if i == 0 else "and "
        title_text += (
            f" {coupler}{v['name']}"
            + f" to {((v['bottom'] * ureg.meters).to(units['depth'])).m:.0f}"
            + f"{units['depth']}"
        )
    title_text += "<br><b>string:</b>"

    last = len(hl.string.sections.keys()) - 1
    for i, (k, v) in enumerate(list(hl.string.sections.items())[::-1]):
        coupler = "" if i == 0 else "and "
        if i == last:
            title_text += (
                f" {coupler}{v['name']}"
                + " to surface"
            )
        else:
            title_text += (
                f" {coupler}{v['name']}"
                + f" ({((v['length'] * ureg.meters).to(units['depth'])).m:.0f}"
                + f"{units['depth']})"
            )

    fig.update_layout(
        title_text=title_text,
        xaxis=dict(
            title=f"Hook-Load ({units['tension']})"
        ),
        yaxis=dict(
            autorange='reversed',
            title=f"MD ({units['depth']})",
            exponentformat='none',
        ),
        annotations=annotations
    )

    return fig



import numpy as np
try:
    from vedo import Circle
    VEDO = True
except ImportError:
    VEDO = False

class Target:
    def __init__(
        self,
        name,
        n,
        e,
        tvd,
        shape,
        locked=0,
        orientation=0,
        dip=0,
        color='green',
        alpha=0.5,
        **geometry
    ):
        """
        Parameters
        ----------
            geometry: 
        """
        assert VEDO, "ImportError: try pip install welleng[easy]"

        SHAPES = [
            'circle',
            'ellipse',
            'rectangle',
            'polygon',
        ]

        GEOMETRY = dict(
            rectangle = ['pos1', 'pos2'],
            circle = ['radius'],
            ellipse = {'radius_1': 0, 'radius_2': 0, 'res': 120},
        )

        assert shape in SHAPES, "shape not in SHAPES"
        assert set(geometry.keys()) == set(GEOMETRY[shape]), 'wrong geometry'

        self.name = name
        self.n = n
        self.e = e
        self.tvd = tvd
        self.shape = shape
        self.locked = locked
        self.orientation = orientation
        self.dip = dip
        self.color = color
        self.alpha = alpha
        self.geometry = geometry

    def plot_data(self):
        pos = [self.n, self.e, self.tvd]
        if self.shape == "circle":
            g = Circle(
                pos=pos,
                r=self.geometry['radius'],
                c=self.color,
                alpha=self.alpha,
                # res=self.geometry['res']
            )
            g.flag = self.name
            g.pos=[self.n, self.e, 0]
            g.rotate(self.dip, point=pos, axis=(0,1,0))
            g.rotate(self.orientation, point=pos, axis=(1,0,0))

        return g



import math
from scipy.optimize import minimize
from scipy import interpolate
try:
    from numba import njit
    NJIT = True
except ImportError:
    NJIT = False
import numpy as np

# Define constants
A = [7.24032, -2.84383e-3, 2.75660e-5]
B = [8.63186, -3.31977e-3, 2.37170e-5]

# Gravitational constant
G = 0.052

# Weighting Material Density in ppg
WEIGHTING_MATERIAL_DENSITY = {
    'barite': 35.,
    'spe_11118': 24.
}


class DensityDiesel:
    def __init__(self):
        """
        An interpolation wrapper of the pressure, temperature and density diesel
        data provided in the SPE 11118 paper.
        """
        psia = np.array([15., 3_000., 7_000., 10_000., 12_500.])
        temp = np.array([100., 200., 300., 350.])

        psia_psia, temp_temp = np.meshgrid(psia, temp)

        rho = np.array([
            [6.9597, 7.0597, 7.1621, 7.2254, 7.2721],
            [6.6598, 6.7690, 6.8789, 6.9464, 6.9930],
            [6.3575, 6.4782, 6.5965, 6.6673, 6.7198],
            [6.2083, 6.3350, 6.4624, 6.5366, 6.5874]
        ])

        self.density_diesel = interpolate.interp2d(
            psia_psia, temp_temp, rho, kind='cubic'
        )

    def get_density(self, pressure, temperature):
        """
        Interpolate diesel density for given pressure and temperature using
        the lookup data provided in SPE 11118 paper.
        """
        density = self.density_diesel(
            pressure, temperature
        )

        return density

class Fluid:
    def __init__(
        self,
        fluid_density,
        reference_temp=32.,
        reference_pressure=0.,
        base_fluid_water_ratio=0.2,
        weighting_material='Barite'
    ):
        """
        Density profile calculated from SPE 11118 Mathematical Field Model
        Predicts Downhold Density Changes in Static Drilling Fluids by Roland
        R. Sorelle et al.

        This paper was written in oilfield units, so we'll convert inputs to
        ppg, ft, F and psi.

        Parameters
        ----------
        fluid_density: float
            The combined fluid density in ppg at reference conditions.
        reference_temp: float (default 32.0)
            The reference temperature in Fahrenheit
        reference_pressure: float (default 0.0)
            The reference pressure in psig.
        weighting_material: str
            The material being used to weight the drilling fluid (see the
            WEIGHTING_MATERIAL_DENSITY dictionary).
        """

        assert weighting_material.lower() in WEIGHTING_MATERIAL_DENSITY.keys()

        self.density_weighting_material = (
            WEIGHTING_MATERIAL_DENSITY.get(weighting_material.lower())
        )
        self.density_fluid_reference = fluid_density
        self.temp_reference = reference_temp
        self.base_fluid_water_ratio = base_fluid_water_ratio
        self.pressure_reference = reference_pressure

        if NJIT:
            self._get_coefficients = njit()(self._get_coefficients)
            self._func = njit()(self._func)
        else:
            self._get_coefficients = self._get_coefficients
            self._func = self._func

        self._get_density_base_fluids()
        self._get_volumes_reference()

    def _get_density_base_fluids(self):
        """
        Equation 1 and 2
        """
        def func(temperature, pressure, c):
            density = c[0] + c[1] * temperature + c[2] * pressure

            return density

        self.density_oil_reference = func(
            self.temp_reference, self.pressure_reference, A
        )
        self.density_water_reference = func(
            self.temp_reference, self.pressure_reference, B
        )

    def _get_volumes_reference(self):
        self.base_fluid_density_reference = (
            self.base_fluid_water_ratio * self.density_water_reference
            + (1 - self.base_fluid_water_ratio) * self.density_oil_reference
        )

        volume_weighting_material = (
            self.density_fluid_reference
            - self.base_fluid_density_reference
        ) / self.density_weighting_material

        volume_total = 1 + volume_weighting_material

        self.volume_water_reference_relative = (
            self.base_fluid_water_ratio / volume_total
        )
        self.volume_oil_reference_relative = (
            (1 - self.base_fluid_water_ratio) / volume_total
        )
        self.volume_weighting_material_relative = (
            volume_weighting_material / volume_total
        )


    @staticmethod
    def _get_coefficients(
        density_average, pressure_applied, temperature_top,
        fluid_thermal_gradient, A0, A1, A2, B0, B1, B2
    ):
        alpha_1 = (
            A0 + A1 * temperature_top + A2 * pressure_applied
        )
        alpha_2 = (
            A1 * fluid_thermal_gradient + G * A2 * density_average
        )
        beta_1 = (
            B0 + B1 * temperature_top + B2 * pressure_applied
        )
        beta_2 = (
            B1 * fluid_thermal_gradient + G * B2 * density_average
        )

        return (alpha_1, alpha_2, beta_1, beta_2)


    @staticmethod
    def _func(
        density_average, density_top, volume_water_relative,
        volume_oil_relative, depth, alpha_1, alpha_2, beta_1, beta_2
    ):
        if depth == 0:
            return density_top
        func = (
            (
                density_top * depth
                - (
                    volume_oil_relative * alpha_1 * density_average
                    / alpha_2
                )
                * math.log(
                    (alpha_1 + alpha_2 * depth) / alpha_1
                )
            ) / (depth * (1 - volume_water_relative - volume_oil_relative))
            - (
                volume_water_relative * beta_1 * density_average / beta_2
                * math.log(
                    (beta_1 + beta_2 * depth) / beta_1
                )
            ) / (depth * (1 - volume_water_relative - volume_oil_relative))
        )
        return func

    def _get_density(
        self, density_average, density_top, temperature_top,
        volume_water_relative, volume_oil_relative, pressure_applied, depth,
        fluid_thermal_gradient
    ):
        density_average = density_average[0]
        alpha_1, alpha_2, beta_1, beta_2 = self._get_coefficients(
            density_average, pressure_applied, temperature_top,
            fluid_thermal_gradient, A[0], A[1], A[2], B[0], B[1], B[2]
        )

        func = self._func(
            density_average, density_top, volume_water_relative,
            volume_oil_relative, depth, alpha_1, alpha_2, beta_1, beta_2
        )

        return abs(density_average - func)

    def get_density_profile(
        self,
        depth,
        temperature,
        pressure_applied=0.,
        density_bounds=(6., 25.)
    ):
        """
        Function that returns a density profile of the fluid, adjusted for
        temperature and compressibility and assuming that the fluid's reference
        parameters are the surface parameters.

        Parameters
        ----------
        depth: float or list or (n) array of floats
            The vertical depth of interest relative to surface in feet.
        temperature: float or list or (n) array of floats
            The temperature corresponding to the vertical depth of interest in
            Fahrenheit.
        pressure_applied: float (default=0.)
            Additional pressure applied to the fluid in psi.
        density_bounds: (2) tuple of floats (default=(6., 25.))
            Density bounds to constrain the optimization algorithm in ppg.
        """

        # Convert to (n) array to manage single float or list/array
        depth = np.array([depth]).reshape(-1)
        with np.errstate(invalid='ignore'):
            temperature_thermal_gradient = np.nan_to_num((
                temperature - self.temp_reference
            ) / depth
            )

        density_profile = [
            minimize(
                fun=self._get_density,
                x0=self.density_fluid_reference,
                args=(
                    self.density_fluid_reference,
                    self.temp_reference,
                    self.volume_water_reference_relative,
                    self.volume_oil_reference_relative,
                    pressure_applied,
                    d,
                    t,
                ),
                method='SLSQP',
                bounds=[density_bounds]
            ).x
            for d, t in zip(depth, temperature_thermal_gradient)
        ]

        return np.vstack(density_profile).reshape(-1).tolist()


def main():
    diesel_density_func = DensityDiesel()
    diesel_density = diesel_density_func.get_density(15, 120)
    """
    An example of initiating a Fluid class and generating a density profile
    for the fluid for a range of depths and temperatures.
    """
    # Define the fluid
    fluid = Fluid(
        fluid_density=10.,  # ppg
        reference_temp=120.,  # Fahrenheit,
        weighting_material='SPE_11118',
        base_fluid_water_ratio=0.103,
    )

    # Override calculated volumes - I can't get the same values as the SPE
    # paper if I build the fluid. However, the fluid properties can be
    # overwritten if desired as indicated below:
    fluid.volume_water_reference_relative = 0.09
    fluid.volume_oil_reference_relative = 0.78
    fluid.volume_weighting_material_relative = 0.11

    depth = np.linspace(0, 10_000, 1001)
    temperature = np.linspace(120, 250, 1001)

    density_profile = fluid.get_density_profile(
        depth=depth,
        temperature=temperature
    )

    # Check we get the same answer as the SPE paper example
    assert round(density_profile[-1], 2) == 9.85

    # load dependencies for plotting results
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    # construct plots
    fig = make_subplots(rows=1, cols=2, shared_yaxes=True)

    fig.add_trace(go.Scatter(
        x=density_profile,
        y=depth,
        mode='lines',
        name='Density (ppg)',

    ), row=1, col=1)
    fig.add_trace(go.Scatter(
        x=temperature,
        y=depth,
        mode='lines',
        name='Temperature (F)'
    ), row=1, col=2)
    fig.update_layout(
        title="Effect of Temperature and Compressibility on Mud Density",
        yaxis=dict(
            autorange='reversed',
            title="TVD (ft)",
            tickformat=",.0f"
        ),
        showlegend=False
    )
    fig.update_xaxes(
        title_text="Density (ppg)",
        tickformat=".2f",
        row=1, col=1
    )
    fig.update_xaxes(
        title_text="Temperature (\xb0F)",
        tickformat=".0f",
        row=1, col=2
    )
    fig.show()


if __name__ == '__main__':
    main()
